#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2018 4AM
#
# SPDX-License-Identifier: MIT

# (c) 2018 by 4am
# MIT-licensed

import argparse
import collections
import json
import os
import sys

__version__ = "1.0"
__date__ = "2018-09-08"
__progname__ = "a2rchery"
__displayname__ = __progname__ + " by 4am (" + __date__ + ")"

# chunk IDs for .a2r files
kA2R2 = b"A2R2"
kINFO = b"INFO"
kSTRM = b"STRM"
kMETA = b"META"
# other things defined in the .a2r specification
kLanguages = ("English","Spanish","French","German","Chinese","Japanese","Italian","Dutch","Portuguese","Danish","Finnish","Norwegian","Swedish","Russian","Polish","Turkish","Arabic","Thai","Czech","Hungarian","Catalan","Croatian","Greek","Hebrew","Romanian","Slovak","Ukrainian","Indonesian","Malay","Vietnamese","Other")
kRequiresRAM = ("16K","24K","32K","48K","64K","128K","256K","512K","768K","1M","1.25M","1.5M+","Unknown")
kRequiresMachine = ("2","2+","2e","2c","2e+","2gs","2c+","3","3+")
kCaptureTiming = 1
kCaptureBits = 2
kCaptureXTiming = 3

# strings and things, for print routines and error messages
sEOF = "Unexpected EOF"
sBadChunkSize = "Bad chunk size"
dNoYes = {False:"no",True:"yes"}
tQuarters = (".00",".25",".50",".75")
dTiming = {kCaptureTiming:"timing",kCaptureBits:"bits",kCaptureXTiming:"xtiming"}

# errors that may be raised
class A2RError(Exception): pass # base class
class A2REOFError(A2RError): pass
class A2RFormatError(A2RError): pass
class A2RHeaderError(A2RError): pass
class A2RHeaderError_NoA2R2(A2RHeaderError): pass
class A2RHeaderError_NoFF(A2RHeaderError): pass
class A2RHeaderError_NoLF(A2RHeaderError): pass
class A2RINFOFormatError(A2RFormatError): pass
class A2RINFOFormatError_BadVersion(A2RINFOFormatError): pass
class A2RINFOFormatError_BadDiskType(A2RINFOFormatError): pass
class A2RINFOFormatError_BadWriteProtected(A2RINFOFormatError): pass
class A2RINFOFormatError_BadSynchronized(A2RINFOFormatError): pass
class A2RINFOFormatError_BadCleaned(A2RINFOFormatError): pass
class A2RINFOFormatError_BadCreator(A2RINFOFormatError): pass
class A2RSTRMFormatError(A2RFormatError): pass
class A2RMETAFormatError(A2RFormatError): pass
class A2RMETAFormatError_DuplicateKey(A2RFormatError): pass
class A2RMETAFormatError_BadValue(A2RFormatError): pass
class A2RMETAFormatError_BadLanguage(A2RFormatError): pass
class A2RMETAFormatError_BadRAM(A2RFormatError): pass
class A2RMETAFormatError_BadMachine(A2RFormatError): pass

class A2RParseError(A2RError):
    pass

def from_uint32(b):
    return int.from_bytes(b, byteorder="little")
from_uint16=from_uint32

def to_uint32(b):
    return b.to_bytes(4, byteorder="little")

def to_uint16(b):
    return b.to_bytes(2, byteorder="little")

def to_uint8(b):
    return b.to_bytes(1, byteorder="little")

def raise_if(cond, e, s=""):
    if cond: raise e(s)

class DiskImage: # base class
    def __init__(self, filename=None, stream=None):
        raise_if(not filename and not stream, A2RError, "no input")
        self.filename = filename
        self.tracks = []

    def seek(self, track_num):
        """returns Track object for the given track, or None if the track is not part of this disk image. track_num can be 0..40 in 0.25 increments (0, 0.25, 0.5, 0.75, 1, &c.)"""
        return None

class A2RValidator:
    def validate_info_version(self, version):
        raise_if(version != b'\x01', A2RINFOFormatError_BadVersion, "Unknown version (expected 1, found %s)" % version)

    def validate_info_disk_type(self, disk_type):
        raise_if(disk_type not in (b'\x01',b'\x02'), A2RINFOFormatError_BadDiskType, "Unknown disk type (expected 1 or 2, found %s)" % disk_type)

    def validate_info_write_protected(self, write_protected):
        raise_if(write_protected not in (b'\x00',b'\x01'), A2RINFOFormatError_BadWriteProtected, "Unknown write protected flag (expected 0 or 1, found %s)" % write_protected)

    def validate_info_synchronized(self, synchronized):
        raise_if(synchronized not in (b'\x00',b'\x01'), A2RINFOFormatError_BadSynchronized, "Unknown synchronized flag (expected 0, or 1, found %s)" % synchronized)

    def validate_info_creator(self, creator_as_bytes):
        raise_if(len(creator_as_bytes) > 32, A2RINFOFormatError_BadCreator, "Creator is longer than 32 bytes")
        try:
            creator_as_bytes.decode("UTF-8")
        except:
            raise_if(True, A2RINFOFormatError_BadCreator, "Creator is not valid UTF-8")

    def encode_info_creator(self, creator_as_string):
        creator_as_bytes = creator_as_string.encode("UTF-8").ljust(32, b" ")
        self.validate_info_creator(creator_as_bytes)
        return creator_as_bytes

    def decode_info_creator(self, creator_as_bytes):
        self.validate_info_creator(creator_as_bytes)
        return creator_as_bytes.decode("UTF-8").strip()

    def validate_metadata(self, metadata_as_bytes):
        try:
            metadata = metadata_as_bytes.decode("UTF-8")
        except:
            raise A2RMETAFormatError("Metadata is not valid UTF-8")

    def decode_metadata(self, metadata_as_bytes):
        self.validate_metadata(metadata_as_bytes)
        return metadata_as_bytes.decode("UTF-8")

    def validate_metadata_value(self, value):
        raise_if("\t" in value, A2RMETAFormatError_BadValue, "Invalid metadata value (contains tab character)")
        raise_if("\n" in value, A2RMETAFormatError_BadValue, "Invalid metadata value (contains linefeed character)")
        raise_if("|" in value, A2RMETAFormatError_BadValue, "Invalid metadata value (contains pipe character)")

    def validate_metadata_language(self, language):
        raise_if(language and (language not in kLanguages), A2RMETAFormatError_BadLanguage, "Invalid metadata language")

    def validate_metadata_requires_ram(self, requires_ram):
        raise_if(requires_ram and (requires_ram not in kRequiresRAM), A2RMETAFormatError_BadRAM, "Invalid metadata requires_ram")

    def validate_metadata_requires_machine(self, requires_machine):
        raise_if(requires_machine and (requires_machine not in kRequiresMachine), A2RMETAFormatError_BadMachine, "Invalid metadata requires_machine")

class A2RReader(DiskImage, A2RValidator):
    def __init__(self, filename=None, stream=None):
        DiskImage.__init__(self, filename, stream)
        self.info = collections.OrderedDict()
        self.meta = collections.OrderedDict()
        self.flux = collections.OrderedDict()

        with stream or open(filename, "rb") as f:
            header_raw = f.read(8)
            raise_if(len(header_raw) != 8, A2REOFError, sEOF)
            self.__process_header(header_raw)
            while True:
                chunk_id = f.read(4)
                if not chunk_id: break
                raise_if(len(chunk_id) != 4, A2REOFError, sEOF)
                chunk_size_raw = f.read(4)
                raise_if(len(chunk_size_raw) != 4, A2REOFError, sEOF)
                chunk_size = from_uint32(chunk_size_raw)
                data = f.read(chunk_size)
                raise_if(len(data) != chunk_size, A2REOFError, sEOF)
                if chunk_id == kINFO:
                    raise_if(chunk_size != 36, A2RFormatError, sBadChunkSize)
                    self.__process_info(data)
                elif chunk_id == kSTRM:
                    self.__process_strm(data)
                elif chunk_id == kMETA:
                    self.__process_meta(data)

    def __process_header(self, data):
        raise_if(data[:4] != kA2R2, A2RHeaderError_NoA2R2, "Magic string 'A2R2' not present at offset 0")
        raise_if(data[4] != 0xFF, A2RHeaderError_NoFF, "Magic byte 0xFF not present at offset 4")
        raise_if(data[5:8] != b"\x0A\x0D\x0A", A2RHeaderError_NoLF, "Magic bytes 0x0A0D0A not present at offset 5")

    def __process_info(self, data):
        version = data[0]
        self.validate_info_version(to_uint8(version))
        disk_type = data[33]
        self.validate_info_disk_type(to_uint8(disk_type))
        write_protected = data[34]
        self.validate_info_write_protected(to_uint8(write_protected))
        synchronized = data[35]
        self.validate_info_synchronized(to_uint8(synchronized))
        creator = self.decode_info_creator(data[1:33])
        self.info["version"] = version # int
        self.info["disk_type"] = disk_type # int
        self.info["write_protected"] = (write_protected == 1) # boolean
        self.info["synchronized"] = (synchronized == 1) # boolean
        self.info["creator"] = creator # string

    def __process_strm(self, data):
        raise_if(data[-1] != 0xFF, A2RSTRMFormatError, "Missing phase reset at end of STRM chunk")
        i = 0
        while i < len(data) - 1:
            location = data[i]
            capture_type = data[i+1]
            data_length = from_uint32(data[i+2:i+6])
            tick_count = from_uint32(data[i+6:i+10])
            if location not in self.flux:
                self.flux[location] = []
            self.flux[location].append(
                {"capture_type": capture_type,
                 "data_length": data_length,
                 "tick_count": tick_count,
                 "data": data[i+10:i+10+data_length]}
            )
            i = i + 10 + data_length

    def __process_meta(self, metadata_as_bytes):
        metadata = self.decode_metadata(metadata_as_bytes)
        for line in metadata.split("\n"):
            if not line: continue
            columns_raw = line.split("\t")
            raise_if(len(columns_raw) != 2, A2RMETAFormatError, "Malformed metadata")
            key, value_raw = columns_raw
            raise_if(key in self.meta, A2RMETAFormatError_DuplicateKey, "Duplicate metadata key %s" % key)
            values = value_raw.split("|")
            if key == "language":
                list(map(self.validate_metadata_language, values))
            elif key == "requires_ram":
                list(map(self.validate_metadata_requires_ram, values))
            elif key == "requires_machine":
                list(map(self.validate_metadata_requires_machine, values))
            self.meta[key] = len(values) == 1 and values[0] or tuple(values)

    def to_json(self):
        j = {"a2r": {"info":self.info, "meta":self.meta}}
        return json.dumps(j, indent=2)

class A2RWriter(A2RValidator):
    def __init__(self, creator):
        self.info = collections.OrderedDict()
        self.meta = collections.OrderedDict()
        self.flux = collections.OrderedDict()

    def from_json(self, json_string):
        j = json.loads(json_string)
        root = [x for x in j.keys()].pop()
        self.meta.update(j[root]["meta"])

    def build_head(self):
        chunk = bytearray()
        chunk.extend(kA2R2) # magic bytes
        chunk.extend(b"\xFF\x0A\x0D\x0A") # more magic bytes
        return chunk

    def build_info(self):
        chunk = bytearray()
        chunk.extend(kINFO) # chunk ID
        chunk.extend(to_uint32(36)) # chunk size (constant)
        version_raw = to_uint8(self.info["version"])
        self.validate_info_version(version_raw)
        creator_raw = self.encode_info_creator(self.info["creator"])
        disk_type_raw = to_uint8(self.info["disk_type"])
        self.validate_info_disk_type(disk_type_raw)
        write_protected_raw = to_uint8(self.info["write_protected"])
        self.validate_info_write_protected(write_protected_raw)
        synchronized_raw = to_uint8(self.info["synchronized"])
        self.validate_info_synchronized(synchronized_raw)
        chunk.extend(version_raw) # version (int, probably 1)
        chunk.extend(creator_raw) # creator
        chunk.extend(disk_type_raw) # disk type (1=5.25 inch, 2=3.5 inch)
        chunk.extend(write_protected_raw) # write-protected (0=no, 1=yes)
        chunk.extend(synchronized_raw) # tracks synchronized (0=no, 1=yes)
        return chunk

    def build_strm(self):
        data_raw = bytearray()
        for location in self.flux.keys():
            for capture in self.flux[location]:
                data_raw.extend(to_uint8(location)) # track where this capture happened
                data_raw.extend(to_uint8(capture["capture_type"])) # 1 = timing, 2 = bits, 3 = xtiming
                data_raw.extend(to_uint32(len(capture["data"]))) # data length in bytes
                data_raw.extend(to_uint32(capture["tick_count"])) # estimated loop point in ticks
                data_raw.extend(capture["data"])
        data_raw.extend(b"\xFF")
        chunk = bytearray()
        chunk.extend(kSTRM) # chunk ID
        chunk.extend(to_uint32(len(data_raw))) # chunk size
        chunk.extend(data_raw) # all stream data
        return chunk

    def build_meta(self):
        if not self.meta: return b""
        meta_tmp = {}
        for key, value_raw in self.meta.items():
            if type(value_raw) == str:
                values = [value_raw]
            else:
                values = value_raw
            meta_tmp[key] = values
            list(map(self.validate_metadata_value, values))
            if key == "language":
                list(map(self.validate_metadata_language, values))
            elif key == "requires_ram":
                list(map(self.validate_metadata_requires_ram, values))
            elif key == "requires_machine":
                list(map(self.validate_metadata_requires_machine, values))
        data = b"\x0A".join(
            [k.encode("UTF-8") + \
             b"\x09" + \
             "|".join(v).encode("UTF-8") \
             for k, v in meta_tmp.items()]) + b"\x0A"
        chunk = bytearray()
        chunk.extend(kMETA) # chunk ID
        chunk.extend(to_uint32(len(data))) # chunk size
        chunk.extend(data)
        return chunk

    def write(self, stream):
        stream.write(self.build_head())
        stream.write(self.build_info())
        stream.write(self.build_strm())
        stream.write(self.build_meta())

#---------- command line interface ----------

class BaseCommand:
    def __init__(self, name):
        self.name = name

    def setup(self, subparser, description=None, epilog=None, help=".a2r disk image", formatter_class=argparse.HelpFormatter):
        self.parser = subparser.add_parser(self.name, description=description, epilog=epilog, formatter_class=formatter_class)
        self.parser.add_argument("file", help=help)
        self.parser.set_defaults(action=self)

    def __call__(self, args):
        self.a2r_image = A2RReader(args.file)

class CommandVerify(BaseCommand):
    def __init__(self):
        BaseCommand.__init__(self, "verify")

    def setup(self, subparser):
        BaseCommand.setup(self, subparser,
                          description="Verify file structure and metadata of a .a2r disk image (produces no output unless a problem is found)")

class CommandDump(BaseCommand):
    kWidth = 30

    def __init__(self):
        BaseCommand.__init__(self, "dump")

    def setup(self, subparser):
        BaseCommand.setup(self, subparser,
                          description="Print all available information and metadata in a .a2r disk image")

    def __call__(self, args):
        BaseCommand.__call__(self, args)
        self.print_flux()
        self.print_meta()
        self.print_info()

    def print_info(self):
        print("INFO:  Format version:".ljust(self.kWidth), "%d" % self.a2r_image.info["version"])
        print("INFO:  Disk type:".ljust(self.kWidth),           ("5.25-inch", "3.5-inch")[self.a2r_image.info["disk_type"]-1])
        print("INFO:  Write protected:".ljust(self.kWidth),     dNoYes[self.a2r_image.info["write_protected"]])
        print("INFO:  Track synchronized:".ljust(self.kWidth),  dNoYes[self.a2r_image.info["synchronized"]])
        print("INFO:  Creator:".ljust(self.kWidth),             self.a2r_image.info["creator"])

    def print_flux(self):
        for location in self.a2r_image.flux:
            for flux_record in self.a2r_image.flux[location]:
                print(("STRM:  Track %d%s" % (location/4, tQuarters[location%4])).ljust(self.kWidth),
                      dTiming[flux_record["capture_type"]], "capture,",
                      flux_record["tick_count"], "ticks")

    def print_meta(self):
        if not self.a2r_image.meta: return
        for key, values in self.a2r_image.meta.items():
            if type(values) == str:
                values = [values]
            print(("META:  " + key + ":").ljust(self.kWidth), values[0])
            for value in values[1:]:
                print("META:  ".ljust(self.kWidth), value)

class CommandExport(BaseCommand):
    def __init__(self):
        BaseCommand.__init__(self, "export")

    def setup(self, subparser):
        BaseCommand.setup(self, subparser,
                          description="Export (as JSON) all information and metadata from a .a2r disk image")

    def __call__(self, args):
        BaseCommand.__call__(self, args)
        print(self.a2r_image.to_json())

class WriterBaseCommand(BaseCommand):
    def __call__(self, args):
        BaseCommand.__call__(self, args)
        self.args = args
        # maintain creator if there is one, otherwise use default
        self.output = A2RWriter(self.a2r_image.info.get("creator", __displayname__))
        self.output.flux = self.a2r_image.flux.copy()
        self.output.info = self.a2r_image.info.copy()
        self.output.meta = self.a2r_image.meta.copy()
        self.update()
        tmpfile = args.file + ".chery"
        with open(tmpfile, "wb") as f:
            self.output.write(f)
        os.rename(tmpfile, args.file)

class CommandEdit(WriterBaseCommand):
    def __init__(self):
        WriterBaseCommand.__init__(self, "edit")

    def setup(self, subparser):
        WriterBaseCommand.setup(self,
                                subparser,
                                description="Edit information and metadata in a .a2r disk image",
                                epilog="""Tips:

 - Use repeated flags to edit multiple fields at once.
 - Use "key:" with no value to delete a metadata field.
 - Keys are case-sensitive.
 - Some values have format restrictions; read the .a2r specification.""",
                                help=".a2r disk image (modified in place)",
                                formatter_class=argparse.RawDescriptionHelpFormatter)
        self.parser.add_argument("-i", "--info", type=str, action="append",
                                 help="""change information field.
INFO format is "key:value".
Acceptable keys are disk_type, write_protected, synchronized, creator, version.
Other keys are ignored.
For boolean fields, use "1" or "true" or "yes" for true, "0" or "false" or "no" for false.""")
        self.parser.add_argument("-m", "--meta", type=str, action="append",
                                 help="""change metadata field.
META format is "key:value".
Standard keys are title, subtitle, publisher, developer, copyright, version, language, requires_ram,
requires_machine, notes, side, side_name, contributor, image_date. Other keys are allowed.""")

    def update(self):
        # add all new info fields
        for i in self.args.info or ():
            k, v = i.split(":", 1)
            if k in ("write_protected","synchronized"):
                v = v.lower() in ("1", "true", "yes")
            self.output.info[k] = v
        # add all new metadata fields, and delete empty ones
        for m in self.args.meta or ():
            k, v = m.split(":", 1)
            v = v.split("|")
            if len(v) == 1:
                v = v[0]
            if v:
                self.output.meta[k] = v
            elif k in self.output.meta.keys():
                del self.output.meta[k]

class CommandImport(WriterBaseCommand):
    def __init__(self):
        WriterBaseCommand.__init__(self, "import")

    def setup(self, subparser):
        WriterBaseCommand.setup(self, subparser,
                                description="Import JSON file to update metadata in a .a2r disk image")

    def update(self):
        self.output.from_json(sys.stdin.read())

if __name__ == "__main__":
    import sys
    raise_if = lambda cond, e, s="": cond and sys.exit("%s: %s" % (e.__name__, s))
    cmds = [CommandDump(), CommandVerify(), CommandEdit(), CommandExport(), CommandImport()]
    parser = argparse.ArgumentParser(prog=__progname__,
                                     description="""A multi-purpose tool for manipulating .a2r disk images.

See '""" + __progname__ + """ <command> -h' for help on individual commands.""",
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument("-v", "--version", action="version", version=__displayname__)
    sp = parser.add_subparsers(dest="command", help="command")
    for command in cmds:
        command.setup(sp)
    args = parser.parse_args()
    args.action(args)



# greaseweazle/codec/ibm/fm.py
#
# Written & released by Keir Fraser <keir.xen@gmail.com>
#
# This is free and unencumbered software released into the public domain.
# See the file COPYING for more details, or visit <http://unlicense.org>.

import binascii
import copy, heapq, struct, functools
import itertools as it
from bitarray import bitarray
import crcmod.predefined

from greaseweazle.codec.ibm import mfm
from greaseweazle.track import MasterTrack, RawTrack

default_revs = 2

def sync(dat, clk=0xc7):
    x = 0
    for i in range(8):
        x <<= 1
        x |= (clk >> (7-i)) & 1
        x <<= 1
        x |= (dat >> (7-i)) & 1
    return bytes(struct.pack('>H', x))

sync_prefix = bitarray(endian='big')
sync_prefix.frombytes(b'\xaa\xaa' + sync(0xf8))
sync_prefix = sync_prefix[:16+10]

iam_sync_bytes = sync(0xfc, 0xd7)
iam_sync = bitarray(endian='big')
iam_sync.frombytes(b'\xaa\xaa' + iam_sync_bytes)

crc16 = crcmod.predefined.Crc('crc-ccitt-false')

sec_sz = mfm.sec_sz
IDAM   = mfm.IDAM
DAM    = mfm.DAM
Sector = mfm.Sector
IAM    = mfm.IAM
    
class IBM_FM:

    IAM  = 0xfc
    IDAM = 0xfe
    DAM  = 0xfb
    DDAM = 0xf8

    gap_presync = 6

    gapbyte = 0xff

    def __init__(self, cyl, head):
        self.cyl, self.head = cyl, head
        self.sectors = []
        self.iams = []

    def summary_string(self):
        nsec, nbad = len(self.sectors), self.nr_missing()
        s = "IBM FM (%d/%d sectors)" % (nsec - nbad, nsec)
        return s

    def has_sec(self, sec_id):
        return self.sectors[sec_id].crc == 0

    def nr_missing(self):
        return len(list(filter(lambda x: x.crc != 0, self.sectors)))

    def flux(self, *args, **kwargs):
        return self.raw_track().flux(*args, **kwargs)


    def decode_raw(self, track):
        flux = track.flux()
        flux.cue_at_index()
        raw = RawTrack(time_per_rev = self.time_per_rev,
                       clock = self.clock, data = flux)
        bits, _ = raw.get_all_data()

        areas = []
        idam = None

        ## 1. Calculate offsets within dump
        
        for offs in bits.itersearch(iam_sync):
            offs += 16
            areas.append(IAM(offs, offs+1*16))
            self.has_iam = True

        for offs in bits.itersearch(sync_prefix):
            offs += 16
            if len(bits) < offs+1*16:
                continue
            mark = decode(bits[offs:offs+1*16].tobytes())[0]
            clock = decode(bits[offs-1:offs+1*16-1].tobytes())[0]
            if clock != 0xc7:
                continue
            if mark == IBM_FM.IDAM:
                s, e = offs, offs+7*16
                if len(bits) < e:
                    continue
                b = decode(bits[s:e].tobytes())
                c,h,r,n = struct.unpack(">x4B2x", b)
                crc = crc16.new(b).crcValue
                if idam is not None:
                    areas.append(idam)
                idam = IDAM(s, e, crc, c=c, h=h, r=r, n=n)
            elif mark == IBM_FM.DAM or mark == IBM_FM.DDAM:
                if idam is None or idam.end - offs > 1000:
                    areas.append(DAM(offs, offs+4*16, 0xffff, mark=mark))
                else:
                    sz = 128 << idam.n
                    s, e = offs, offs+(1+sz+2)*16
                    if len(bits) < e:
                        continue
                    b = decode(bits[s:e].tobytes())
                    crc = crc16.new(b).crcValue
                    dam = DAM(s, e, crc, mark=mark, data=b[1:-2])
                    areas.append(Sector(idam, dam))
                idam = None
            else:
                pass #print("Unknown mark %02x" % mark)

        if idam is not None:
            areas.append(idam)

        # Convert to offsets within track
        areas.sort(key=lambda x:x.start)
        index = iter(raw.revolutions)
        p, n = 0, next(index)
        for a in areas:
            if a.start >= n:
                p = n
                try:
                    n = next(index)
                except StopIteration:
                    n = float('inf')
            a.delta(p)
        areas.sort(key=lambda x:x.start)

        # Add to the deduped lists
        for a in areas:
            match = False
            if isinstance(a, IAM):
                list = self.iams
            elif isinstance(a, Sector):
                list = self.sectors
            else:
                continue
            for s in list:
                if abs(s.start - a.start) < 1000:
                    match = True
                    break
            if match and isinstance(a, Sector) and s.crc != 0 and a.crc == 0:
                self.sectors = [x for x in self.sectors if x != a]
                match = False
            if not match:
                list.append(a)


    def raw_track(self):

        areas = heapq.merge(self.iams, self.sectors, key=lambda x:x.start)
        t = bytes()

        for a in areas:
            start = a.start//16 - self.gap_presync
            gap = max(start - len(t)//2, 0)
            t += encode(bytes([self.gapbyte] * gap))
            t += encode(bytes(self.gap_presync))
            if isinstance(a, IAM):
                t += iam_sync_bytes
            elif isinstance(a, Sector):
                idam = bytes([self.IDAM,
                              a.idam.c, a.idam.h, a.idam.r, a.idam.n])
                idam += struct.pack('>H', crc16.new(idam).crcValue)
                t += sync(idam[0]) + encode(idam[1:])
                start = a.dam.start//16 - self.gap_presync
                gap = max(start - len(t)//2, 0)
                t += encode(bytes([self.gapbyte] * gap))
                t += encode(bytes(self.gap_presync))
                dam = bytes([a.dam.mark]) + a.dam.data
                dam += struct.pack('>H', crc16.new(dam).crcValue)
                t += sync(dam[0]) + encode(dam[1:])

        # Add the pre-index gap.
        tlen = int((self.time_per_rev / self.clock) // 16)
        gap = max(tlen - len(t)//2, 0)
        t += encode(bytes([self.gapbyte] * gap))

        track = MasterTrack(
            bits = t,
            time_per_rev = self.time_per_rev)
        track.verify = self
        track.verify_revs = default_revs
        return track


class IBM_FM_Formatted(IBM_FM):

    gap_4a = 40 # Post-Index
    gap_1  = 26 # Post-IAM
    gap_2  = 11 # Post-IDAM

    def __init__(self, cyl, head):

        super().__init__(cyl, head)
        self.raw_iams, self.raw_sectors = [], []

    def decode_raw(self, track):
        iams, sectors = self.iams, self.sectors
        self.iams, self.sectors = self.raw_iams, self.raw_sectors
        super().decode_raw(track)
        self.iams, self.sectors = iams, sectors
        for r in self.raw_sectors:
            if r.idam.crc != 0:
                continue
            for s in self.sectors:
                if (s.idam.c == r.idam.c and
                    s.idam.h == r.idam.h and
                    s.idam.r == r.idam.r and
                    s.idam.n == r.idam.n):
                    s.idam.crc = 0
                    if r.dam.crc == 0 and s.dam.crc != 0:
                        s.dam.crc = s.crc = 0
                        s.dam.data = r.dam.data

    def set_img_track(self, tdat):
        pos = 0
        self.sectors.sort(key = lambda x: x.idam.r)
        totsize = functools.reduce(lambda x, y: x + (128<<y.idam.n),
                                   self.sectors, 0)
        if len(tdat) < totsize:
            tdat += bytes(totsize - len(tdat))
        for s in self.sectors:
            s.crc = s.idam.crc = s.dam.crc = 0
            size = 128 << s.idam.n
            s.dam.data = tdat[pos:pos+size]
            pos += size
        self.sectors.sort(key = lambda x: x.start)
        return totsize

    def get_img_track(self):
        tdat = bytearray()
        sectors = self.sectors.copy()
        sectors.sort(key = lambda x: x.idam.r)
        for s in sectors:
            tdat += s.dam.data
        return tdat
        
    def verify_track(self, flux):
        readback_track = IBM_FM_Formatted(self.cyl, self.head)
        readback_track.clock = self.clock
        readback_track.time_per_rev = self.time_per_rev
        for x in self.iams:
            readback_track.iams.append(copy.copy(x))
        for x in self.sectors:
            idam, dam = copy.copy(x.idam), copy.copy(x.dam)
            idam.crc, dam.crc = 0xffff, 0xffff
            readback_track.sectors.append(Sector(idam, dam))
        readback_track.decode_raw(flux)
        if readback_track.nr_missing() != 0:
            return False
        return self.sectors == readback_track.sectors


class IBM_FM_Predefined(IBM_FM_Formatted):

    cskew = 0
    hskew = 0
    interleave = 1
    
    def __init__(self, cyl, head):

        super().__init__(cyl, head)

        # Create logical sector map in rotational order
        sec_map = [-1] * self.nsec
        pos = (cyl*self.cskew + head*self.hskew) % self.nsec
        for i in range(self.nsec):
            while sec_map[pos] != -1:
                pos = (pos + 1) % self.nsec
            sec_map[pos] = i
            pos = (pos + self.interleave) % self.nsec

        pos = self.gap_4a
        if self.gap_1 is not None:
            self.iams = [IAM(pos*16,(pos+1)*16)]
            pos += 1 + self.gap_1

        for i in range(self.nsec):
            pos += self.gap_presync
            idam = IDAM(pos*16, (pos+7)*16, 0xffff,
                        c=cyl, h=head, r=self.id0+sec_map[i], n = self.sz)
            pos += 7 + self.gap_2 + self.gap_presync
            size = 128 << self.sz
            dam = DAM(pos*16, (pos+1+size+2)*16, 0xffff,
                      mark=self.DAM, data=bytes(size))
            self.sectors.append(Sector(idam, dam))
            pos += 1 + size + 2 + self.gap_3

    @classmethod
    def decode_track(cls, cyl, head, track):
        mfm = cls(cyl, head)
        mfm.decode_raw(track)
        return mfm


class Acorn_DFS(IBM_FM_Predefined):

    time_per_rev = 0.2
    clock = 4e-6

    gap_1  = 0 # No IAM
    gap_3  = 21
    nsec   = 10
    id0    = 0
    sz     = 1
    cskew  = 3


encode_list = []
for x in range(256):
    y = 0
    for i in range(8):
        y <<= 1
        y |= 1
        y <<= 1
        y |= (x >> (7-i)) & 1
    encode_list.append(y)

def encode(dat):
    out = bytearray()
    for x in dat:
        out += struct.pack('>H', encode_list[x])
    return bytes(out)

decode = mfm.decode


# Local variables:
# python-indent: 4
# End:



# greaseweazle/image/image.py
#
# Written & released by Keir Fraser <keir.xen@gmail.com>
#
# This is free and unencumbered software released into the public domain.
# See the file COPYING for more details, or visit <http://unlicense.org>.

import os

from .. import error

class Image:

    read_only = False

    ## Context manager for image objects created using .to_file()

    def __enter__(self):
        self.file = open(self.filename, ('wb','xb')[self.noclobber])
        return self

    def __exit__(self, type, value, tb):
        save = type is None or type is KeyboardInterrupt
        try:
            if save:
                # No error: Normal writeout.
                self.file.write(self.get_image())
        finally:
            # Always close the file.
            self.file.close()
        if not save:
            # An error occurred: We remove the target file.
            os.remove(self.filename)

    ## Default .to_file() constructor
    @classmethod
    def to_file(cls, name, fmt, noclobber):
        error.check(not cls.read_only,
                    "%s: Cannot create %s image files" % (name, cls.__name__))
        obj = cls()
        obj.filename = name
        obj.fmt = fmt
        obj.noclobber = noclobber
        return obj

    # Maximum non-empty cylinder on each head, or -1 if no cylinders exist.
    # Returns a list of integers, indexed by head.
    def max_cylinder(self):
        r = list()
        for h in range(2):
            for c in range(100, -2, -1):
                if c < 0 or self.get_track(c,h) is not None:
                    r.append(c)
                    break
        return r

    ## Above methods and class variables can be overridden by subclasses.
    ## Additionally, subclasses must provide following public interfaces:

    ## Read support:
    # def from_file(cls, name)
    # def get_track(self, cyl, side)

    ## Write support (if not cls.read_only):
    # def emit_track(self, cyl, side, track)
    ## Plus either:
    # def get_image(self)
    ## Or:
    # __enter__ / __exit__


# Local variables:
# python-indent: 4
# End:



# greaseweazle/track.py
#
# Written & released by Keir Fraser <keir.xen@gmail.com>
#
# This is free and unencumbered software released into the public domain.
# See the file COPYING for more details, or visit <http://unlicense.org>.

import binascii
import itertools as it
from bitarray import bitarray
from .flux import Flux, WriteoutFlux

# Precompensation to apply to a MasterTrack for writeout.
class Precomp:
    MFM = 0
    FM  = 1
    GCR = 2
    TYPESTRING = [ 'MFM', 'FM', 'GCR' ]
    def __str__(self):
        return "Precomp: %s, %dns" % (Precomp.TYPESTRING[self.type], self.ns)
    def __init__(self, type, ns):
        self.type = type
        self.ns = ns
    def apply(self, bits, bit_ticks, scale):
        t = self.ns * scale
        if self.type == Precomp.MFM:
            for i in bits.itersearch(bitarray('10100', endian='big')):
                bit_ticks[i+2] -= t
                bit_ticks[i+3] += t
            for i in bits.itersearch(bitarray('00101', endian='big')):
                bit_ticks[i+2] += t
                bit_ticks[i+3] -= t
        # This is primarily for GCR and FM which permit adjacent 1s (and
        # have correspondingly slower bit times). However it may be useful
        # for illegal MFM sequences too, especially on Amiga (custom syncwords,
        # 4us-bitcell tracks). Normal MFM should not trigger these patterns.
        for i in bits.itersearch(bitarray('110', endian='big')):
            bit_ticks[i+1] -= t
            bit_ticks[i+2] += t
        for i in bits.itersearch(bitarray('011', endian='big')):
            bit_ticks[i+1] += t
            bit_ticks[i+2] -= t


# A pristine representation of a track, from a codec and/or a perfect image.
class MasterTrack:

    @property
    def bitrate(self):
        return len(self.bits) / self.time_per_rev

    def scale(self, factor):
        """Scale up index timing by specified factor."""
        self.time_per_rev *= factor

    # bits: Track bitcell data, aligned to the write splice (bitarray or bytes)
    # time_per_rev: Time per revolution, in seconds (float)
    # bit_ticks: Per-bitcell time values, in unitless 'ticks'
    # splice: Location of the track splice, in bitcells, after the index
    # weak: List of (start, length) weak ranges
    def __init__(self, bits, time_per_rev, bit_ticks=None, splice=0, weak=[]):
        if isinstance(bits, bytes):
            self.bits = bitarray(endian='big')
            self.bits.frombytes(bits)
        else:
            self.bits = bits
        self.time_per_rev = time_per_rev
        self.bit_ticks = bit_ticks
        self.splice = splice
        self.weak = weak
        self.precomp = None
        self.force_random_weak = True

    def __str__(self):
        s = "\nMaster Track: splice @ %d\n" % self.splice
        s += (" %d bits, %.1f kbit/s"
              % (len(self.bits), self.bitrate/1000))
        if self.bit_ticks:
            s += " (variable)"
        s += ("\n %.1f ms / rev (%.1f rpm)"
              % (self.time_per_rev * 1000, 60 / self.time_per_rev))
        if len(self.weak) > 0:
            s += "\n %d weak range" % len(self.weak)
            if len(self.weak) > 1: s += "s"
            s += ": " + ", ".join(str(n) for _,n in self.weak) + " bits"
        #s += str(binascii.hexlify(self.bits.tobytes()))
        return s

    def summary_string(self):
        s = ('Bitcells (%d bits, %.1f kbit/s, %.1f rpm'
             % (len(self.bits), self.bitrate/1000, 60 / self.time_per_rev))
        if self.bit_ticks:
            s += ', variable'
        if len(self.weak) > 0:
            s += ', weak'
        s += ')'
        return s

    def flux_for_writeout(self, cue_at_index=True):
        return self.flux(for_writeout=True, cue_at_index=cue_at_index)

    def flux(self, for_writeout=False, cue_at_index=True):

        # We're going to mess with the track data, so take a copy.
        bits = self.bits.copy()
        bitlen = len(bits)

        # Also copy the bit_ticks array (or create a dummy one), and remember
        # the total ticks that it contains.
        bit_ticks = self.bit_ticks.copy() if self.bit_ticks else [1] * bitlen
        ticks_to_index = sum(bit_ticks)

        # Weak regions need special processing for correct flux representation.
        for s,n in self.weak:
            e = s + n
            assert 0 < s < e < bitlen
            pattern = bitarray(endian="big")
            if n < 400 or self.force_random_weak:
                # Short weak regions are written with no flux transitions.
                # Actually we insert a flux transition every 32 bitcells, else
                # we risk triggering Greaseweazle's No Flux Area generator.
                pattern.frombytes(b"\x80\x00\x00\x00")
                bits[s:e] = (pattern * (n//32+1))[:n]
            else:
                # Long weak regions we present a fuzzy clock bit in an
                # otherwise normal byte (16 bits MFM). The byte may be
                # interpreted as
                # MFM 0001001010100101 = 12A5 = byte 0x43, or
                # MFM 0001001010010101 = 1295 = byte 0x47
                pattern.frombytes(b"\x12\xA5")
                bits[s:e] = (pattern * (n//16+1))[:n]
                for i in range(0, n-10, 16):
                    x, y = bit_ticks[s+i+10], bit_ticks[s+i+11]
                    bit_ticks[s+i+10], bit_ticks[s+i+11] = x+y*0.5, y*0.5
            # To prevent corrupting a preceding sync word by effectively
            # starting the weak region early, we start with a 1 if we just
            # clocked out a 0.
            bits[s] = not bits[s-1]
            # Similarly modify the last bit of the weak region.
            bits[e-1] = not(bits[e-2] or bits[e])

        if cue_at_index or not for_writeout:
            # Rotate data to start at the index.
            index = -self.splice % bitlen
            if index != 0:
                bits = bits[index:] + bits[:index]
                bit_ticks = bit_ticks[index:] + bit_ticks[:index]
            splice_at_index = index < 4 or bitlen - index < 4
        else:
            splice_at_index = False

        if not for_writeout:
            # Do not extend the track for reliable writeout to disk.
            pass
        elif not cue_at_index:
            # We write the track wherever it may fall (uncued).
            # We stretch the track with extra header gap bytes, in case the
            # drive spins slow and we need more length to create an overlap.
            # Thus if the drive spins slow, the track gets a longer header.
            pos = 4
            # We stretch by 10 percent, which is way more than enough.
            rep = bitlen // (10 * 32)
            bit_ticks = bit_ticks[pos:pos+32] * rep + bit_ticks[pos:]
            bits = bits[pos:pos+32] * rep + bits[pos:]
        elif splice_at_index:
            # Splice is at the index (or within a few bitcells of it).
            # We stretch the track with extra footer gap bytes, in case the
            # drive motor spins slower than expected and we need more filler
            # to get us to the index pulse (where the write will terminate).
            # Thus if the drive spins slow, the track gets a longer footer.
            pos = (self.splice - 4) % bitlen
            # We stretch by 10 percent, which is way more than enough.
            rep = bitlen // (10 * 32)
            bit_ticks = bit_ticks[:pos] + bit_ticks[pos-32:pos] * rep
            bits = bits[:pos] + bits[pos-32:pos] * rep
        else:
            # Splice is not at the index. We will write more than one
            # revolution, and terminate the second revolution at the splice.
            # For the first revolution we repeat the track header *backwards*
            # to the very start of the write. This is in case the drive motor
            # spins slower than expected and the write ends before the original
            # splice position.
            # Thus if the drive spins slow, the track gets a longer header.
            bit_ticks += bit_ticks[:self.splice-4]
            bits += bits[:self.splice-4]
            pos = self.splice+4
            fill_pattern = bits[pos:pos+32]
            while pos >= 32:
                pos -= 32
                bits[pos:pos+32] = fill_pattern

        if for_writeout and self.precomp is not None:
            self.precomp.apply(bits, bit_ticks,
                               ticks_to_index / (self.time_per_rev*1e9))

        # Convert the stretched track data into flux.
        bit_ticks_i = iter(bit_ticks)
        flux_list = []
        flux_ticks = 0
        for bit in bits:
            flux_ticks += next(bit_ticks_i)
            if bit:
                flux_list.append(flux_ticks)
                flux_ticks = 0
        if flux_ticks and for_writeout:
            flux_list.append(flux_ticks)

        # Package up the flux for return.
        if for_writeout:
            flux = WriteoutFlux(ticks_to_index, flux_list,
                                ticks_to_index / self.time_per_rev,
                                index_cued = cue_at_index,
                                terminate_at_index = splice_at_index)
        else:
            flux = Flux([ticks_to_index]*2, flux_list*2,
                        ticks_to_index / self.time_per_rev,
                        index_cued = True)
            flux.splice = sum(bit_ticks[:self.splice])

        return flux


# Track data generated from flux.
class RawTrack:

    # clock: Expected time per raw bitcell, in seconds (float)
    # data: Flux object, or a form convertible to a Flux object
    # time_per_rev: Expected time per revolution, in seconds (optional, float)
    def __init__(self, clock, data, time_per_rev=None):
        self.clock = clock
        self.time_per_rev = time_per_rev
        self.clock_max_adj = 0.10
        self.pll_period_adj = 0.05
        self.pll_phase_adj = 0.60
        self.bitarray = bitarray(endian='big')
        self.timearray = []
        self.revolutions = []
        self.import_flux_data(data)


    def __str__(self):
        s = "\nRaw Track: %d revolutions\n" % len(self.revolutions)
        for rev in range(len(self.revolutions)):
            b, _ = self.get_revolution(rev)
            s += "Revolution %u (%u bits): " % (rev, len(b))
            s += str(binascii.hexlify(b.tobytes())) + "\n"
        b = self.bitarray[sum(self.revolutions):]
        s += "Tail (%u bits): " % (len(b))
        s += str(binascii.hexlify(b.tobytes())) + "\n"
        return s[:-1]


    def get_revolution(self, nr):
        start = sum(self.revolutions[:nr])
        end = start + self.revolutions[nr]
        return self.bitarray[start:end], self.timearray[start:end]


    def get_all_data(self):
        return self.bitarray, self.timearray


    def import_flux_data(self, data):

        flux = data.flux()
        freq = flux.sample_freq
        if self.time_per_rev is not None:
            # Adjust the raw flux to have the expected time per revolution.
            freq *= flux.time_per_rev / self.time_per_rev

        clock = self.clock
        clock_min = self.clock * (1 - self.clock_max_adj)
        clock_max = self.clock * (1 + self.clock_max_adj)

        index_iter = it.chain(iter(map(lambda x: x/freq, flux.index_list)),
                              [float('inf')])

        # Make sure there's enough time in the flux list to cover all
        # revolutions by appending a "large enough" final flux value.
        tail = max(0, sum(flux.index_list) - sum(flux.list) + clock*freq*2)
        flux_iter = it.chain(flux.list, [tail])

        flux_to_bitcells(
            self.bitarray, self.timearray, self.revolutions,
            index_iter, flux_iter,
            freq, clock, clock_min, clock_max,
            self.pll_period_adj, self.pll_phase_adj)

            
def flux_to_bitcells(bit_array, time_array, revolutions,
                     index_iter, flux_iter,
                     freq, clock_centre, clock_min, clock_max,
                     pll_period_adj, pll_phase_adj):

    nbits = 0
    ticks = 0.0
    clock = clock_centre
    to_index = next(index_iter)

    for x in flux_iter:

        # Gather enough ticks to generate at least one bitcell.
        ticks += x / freq
        if ticks < clock/2:
            continue

        # Clock out zero or more 0s, followed by a 1.
        zeros = 0
        while True:

            # Check if we cross the index mark.
            to_index -= clock
            if to_index < 0:
                revolutions.append(nbits)
                nbits = 0
                to_index += next(index_iter)

            nbits += 1
            ticks -= clock
            time_array.append(clock)
            if ticks >= clock/2:
                zeros += 1
                bit_array.append(False)
            else:
                bit_array.append(True)
                break

        # PLL: Adjust clock frequency according to phase mismatch.
        if zeros <= 3:
            # In sync: adjust clock by a fraction of the phase mismatch.
            clock += ticks * pll_period_adj
        else:
            # Out of sync: adjust clock towards centre.
            clock += (clock_centre - clock) * pll_period_adj
        # Clamp the clock's adjustment range.
        clock = min(max(clock, clock_min), clock_max)
        # PLL: Adjust clock phase according to mismatch.
        new_ticks = ticks * (1 - pll_phase_adj)
        time_array[-1] += ticks - new_ticks
        ticks = new_ticks

# Local variables:
# python-indent: 4
# End:



# greaseweazle/image/acorn.py
#
# Written & released by Keir Fraser <keir.xen@gmail.com>
#
# This is free and unencumbered software released into the public domain.
# See the file COPYING for more details, or visit <http://unlicense.org>.

from .image.img import IMG

class SSD(IMG):
    default_format = 'acorn.dfs.ss'

class DSD(IMG):
    default_format = 'acorn.dfs.ds'

class ADS(IMG):
    default_format = 'acorn.adfs.160'
    
class ADM(IMG):
    default_format = 'acorn.adfs.320'
    
class ADL(IMG):
    default_format = 'acorn.adfs.640'

# Local variables:
# python-indent: 4
# End:



# greaseweazle/image/sf7.py
#
# Written & released by Keir Fraser <keir.xen@gmail.com>
#
# This is free and unencumbered software released into the public domain.
# See the file COPYING for more details, or visit <http://unlicense.org>.

from ..image.img import IMG

class SF7(IMG):
    default_format = 'sega.sf7000'

# Local variables:
# python-indent: 4
# End:



# greaseweazle/image/adf.py
#
# Written & released by Keir Fraser <keir.xen@gmail.com>
#
# This is free and unencumbered software released into the public domain.
# See the file COPYING for more details, or visit <http://unlicense.org>.

from . import error
from . import codec.amiga.amigados as amigados
from .img import IMG
from .image import Image

from ..codec import formats

class ADF(Image):

    default_format = 'amiga.amigados'

    def __init__(self, name, fmt):
        self.to_track = dict()
        error.check(fmt is not None and fmt.adf_compatible, """\
ADF image requires compatible format conversion
Compatible formats:\n%s"""
                    % formats.print_formats(lambda k, v: v.adf_compatible))
        self.filename = name
        self.fmt = fmt


    @classmethod
    def from_file(cls, name, fmt):

        if fmt is not None and fmt.img_compatible: # Acorn ADF
            return IMG.from_file(name, fmt)

        with open(name, "rb") as f:
            dat = f.read()

        adf = cls(name, fmt)

        while True:
            nsec = fmt.fmt.nsec
            error.check((len(dat) % (2*nsec*512)) == 0, "Bad ADF image length")
            ncyl = len(dat) // (2*nsec*512)
            if ncyl < 90:
                break
            error.check(nsec == 11, "Bad ADF image length")
            fmt = adf.fmt = formats.Format_Amiga_AmigaDOS_HD()

        pos = 0
        for t in fmt.max_tracks:
            tnr = t.cyl*2 + t.head
            ados = fmt.fmt(t.cyl, t.head)
            pos += ados.set_adf_track(dat[pos:])
            adf.to_track[tnr] = ados

        return adf


    @classmethod
    def to_file(cls, name, fmt, noclobber):
        if fmt is not None and fmt.img_compatible: # Acorn ADF
            return IMG.to_file(name, fmt, noclobber)
        adf = cls(name, fmt)
        adf.noclobber = noclobber
        return adf


    def get_track(self, cyl, side):
        tnr = cyl * 2 + side
        if not tnr in self.to_track:
            return None
        return self.to_track[tnr].raw_track()


    def emit_track(self, cyl, side, track):
        tnr = cyl * 2 + side
        self.to_track[tnr] = track


    def get_image(self):

        tdat = bytearray()

        ntracks = max(self.to_track, default=0) + 1

        for tnr in range(ntracks):
            t = self.to_track[tnr] if tnr in self.to_track else None
            if t is not None and hasattr(t, 'get_adf_track'):
                tdat += t.get_adf_track()
            elif tnr < 160:
                # Pad empty/damaged tracks.
                tdat += amigados.bad_sector * self.fmt.fmt.nsec
            else:
                # Do not extend past 160 tracks unless there is data.
                break

        if ntracks < 160:
            tdat += amigados.bad_sector * self.fmt.fmt.nsec * (160 - ntracks)

        return tdat


# Local variables:
# python-indent: 4
# End:



# greaseweazle/image/ipf.py
#
# Written & released by Keir Fraser <keir.xen@gmail.com>
# 
# This is free and unencumbered software released into the public domain.
# See the file COPYING for more details, or visit <http://unlicense.org>.

import os, sys
import platform
import ctypes as ct
import itertools as it
from bitarray import bitarray
from ..track import MasterTrack, RawTrack
from .. import error
from .image import Image

class CapsDateTimeExt(ct.Structure):
    _pack_ = 1
    _fields_ = [
        ('year',  ct.c_uint),
        ('month', ct.c_uint),
        ('day',   ct.c_uint),
        ('hour',  ct.c_uint),
        ('min',   ct.c_uint),
        ('sec',   ct.c_uint),
        ('tick',  ct.c_uint)]

class CapsImageInfo(ct.Structure):
    platform_name = [
        "N/A", "Amiga", "Atari ST", "IBM PC", "Amstrad CPC",
        "Spectrum", "Sam Coupe", "Archimedes", "C64", "Atari (8-bit)" ]
    _pack_ = 1
    _fields_ = [
        ('type',        ct.c_uint), # image type
        ('release',     ct.c_uint), # release ID
        ('revision',    ct.c_uint), # release revision ID
        ('mincylinder', ct.c_uint), # lowest cylinder number
        ('maxcylinder', ct.c_uint), # highest cylinder number
        ('minhead',     ct.c_uint), # lowest head number
        ('maxhead',     ct.c_uint), # highest head number
        ('crdt',        CapsDateTimeExt), # image creation date.time
        ('platform',    ct.c_uint * 4)] # intended platform(s)

class CapsTrackInfoT2(ct.Structure):
    _pack_ = 1
    _fields_ = [
        ('type',       ct.c_uint), # track type
        ('cylinder',   ct.c_uint), # cylinder#
        ('head',       ct.c_uint), # head#
        ('sectorcnt',  ct.c_uint), # available sectors
        ('sectorsize', ct.c_uint), # sector size, unused
        ('trackbuf',   ct.POINTER(ct.c_ubyte)), # track buffer memory 
        ('tracklen',   ct.c_uint), # track buffer memory length
        ('timelen',    ct.c_uint), # timing buffer length
        ('timebuf',    ct.POINTER(ct.c_uint)), # timing buffer
        ('overlap',    ct.c_int),  # overlap position
        ('startbit',   ct.c_uint), # start position of the decoding
        ('wseed',      ct.c_uint), # weak bit generator data
        ('weakcnt',    ct.c_uint)] # number of weak data areas

class CapsSectorInfo(ct.Structure):
    _pack_ = 1
    _fields_ = [
        ('descdatasize', ct.c_uint), # data size in bits from IPF descriptor
        ('descgapsize', ct.c_uint),  # gap size in bits from IPF descriptor
        ('datasize', ct.c_uint),     # data size in bits from decoder
        ('gapsize', ct.c_uint),      # gap size in bits from decoder
        ('datastart', ct.c_uint),    # data start pos in bits from decoder
        ('gapstart', ct.c_uint),     # gap start pos in bits from decoder
        ('gapsizews0', ct.c_uint),   # gap size before write splice
        ('gapsizews1', ct.c_uint),   # gap size after write splice
        ('gapws0mode', ct.c_uint),   # gap size mode before write splice
        ('gapws1mode', ct.c_uint),   # gap size mode after write splice
        ('celltype', ct.c_uint),     # bitcell type
        ('enctype', ct.c_uint)]      # encoder type

class CapsDataInfo(ct.Structure):
    _pack_ = 1
    _fields_ = [
        ('type', ct.c_uint),  # data type
        ('start', ct.c_uint), # start position
        ('size', ct.c_uint)]  # size in bits

class DI_LOCK:
    INDEX     = 1<<0
    ALIGN     = 1<<1
    DENVAR    = 1<<2
    DENAUTO   = 1<<3
    DENNOISE  = 1<<4
    NOISE     = 1<<5
    NOISEREV  = 1<<6
    MEMREF    = 1<<7
    UPDATEFD  = 1<<8
    TYPE      = 1<<9
    DENALT    = 1<<10
    OVLBIT    = 1<<11
    TRKBIT    = 1<<12
    NOUPDATE  = 1<<13
    SETWSEED  = 1<<14
    def_flags = (DENVAR | UPDATEFD | TYPE | OVLBIT | TRKBIT)

class IPFTrack(MasterTrack):

    verify_revs = 2
    tolerance = 100

    @staticmethod
    def strong_data(sector, weak):
        """Return list of sector data areas excluding weak sections."""
        def range_next(i):
            s,l = next(i)
            return s, s+l
        weak_tol = 16 # Skip this number of bits after a weak area
        weak_iter = it.chain(weak, [(1<<30,1)])
        ws,we = -1,-1
        sector_iter = iter(sector)
        s,e = range_next(sector_iter)
        try:
            while True:
                while we <= s:
                    ws,we = range_next(weak_iter)
                    we += weak_tol
                if ws < e:
                    if s < ws:
                        yield (s,ws-s)
                    s = we
                else:
                    yield (s,e-s)
                    s = e
                if s >= e:
                    s,e = range_next(sector_iter)
        except StopIteration:
            pass

    def verify_track(self, flux):
        flux.cue_at_index()
        raw = RawTrack(clock = self.time_per_rev/len(self.bits), data = flux)
        raw_bits, _ = raw.get_all_data()
        for s,l in IPFTrack.strong_data(self.sectors, self.weak):
            sector = self.bits[s:s+l]
            # Search within an area +/- the pre-defined # bitcells tolerance
            raw_area = raw_bits[max(self.splice + s - self.tolerance, 0)
                                : self.splice + s + l + self.tolerance]
            # All we care about is at least one match (this is a bit fuzzy)
            if next(raw_area.itersearch(sector), None) is None:
                return False
        return True

class IPF(Image):

    read_only = True

    def __init__(self):
        self.lib = get_libcaps()

    def __del__(self):
        try:
            self.lib.CAPSUnlockAllTracks(self.iid)
            self.lib.CAPSUnlockImage(self.iid)
            self.lib.CAPSRemImage(self.iid)
            del(self.iid)
        except AttributeError:
            pass

    def __str__(self):
        pi = self.pi
        s = "IPF Image File:"
        s += "\n SPS ID: %04d (rev %d)" % (pi.release, pi.revision)
        s += "\n Platform: "
        nr_platforms = 0
        for p in pi.platform:
            if p == 0 and nr_platforms != 0:
                break
            if nr_platforms > 0:
                s += ", "
            s += pi.platform_name[p]
            nr_platforms += 1
        s += ("\n Created: %d/%d/%d %02d:%02d:%02d"
             % (pi.crdt.year, pi.crdt.month, pi.crdt.day,
                pi.crdt.hour, pi.crdt.min, pi.crdt.sec))
        s += ("\n Cyls: %d-%d  Heads: %d-%d"
              % (pi.mincylinder, pi.maxcylinder, pi.minhead, pi.maxhead))
        return s

    @classmethod
    def from_file(cls, name):

        ipf = cls()

        ipf.iid = ipf.lib.CAPSAddImage()
        error.check(ipf.iid >= 0, "Could not create IPF image container")
        cname = ct.c_char_p(name.encode())
        res = ipf.lib.CAPSLockImage(ipf.iid, cname)
        error.check(res == 0, "Could not open IPF image '%s'" % name)
        res = ipf.lib.CAPSLoadImage(ipf.iid, DI_LOCK.def_flags)
        error.check(res == 0, "Could not load IPF image '%s'" % name)
        ipf.pi = CapsImageInfo()
        res = ipf.lib.CAPSGetImageInfo(ct.byref(ipf.pi), ipf.iid)
        error.check(res == 0, "Could not get info for IPF '%s'" % name)
        print(ipf)

        return ipf


    def get_track(self, cyl, head):
        pi = self.pi
        if head < pi.minhead or head > pi.maxhead:
            return None
        if cyl < pi.mincylinder or cyl > pi.maxcylinder:
            return None

        ti = CapsTrackInfoT2(2)
        res = self.lib.CAPSLockTrack(ct.byref(ti), self.iid,
                                     cyl, head, DI_LOCK.def_flags)
        error.check(res == 0, "Could not lock IPF track %d.%d" % (cyl, head))

        if not ti.trackbuf:
            return None # unformatted/empty
        carray_type = ct.c_ubyte * ((ti.tracklen+7)//8)
        carray = carray_type.from_address(
            ct.addressof(ti.trackbuf.contents))
        trackbuf = bitarray(endian='big')
        trackbuf.frombytes(bytes(carray))
        trackbuf = trackbuf[:ti.tracklen]

        data = []
        for i in range(ti.sectorcnt):
            si = CapsSectorInfo()
            res = self.lib.CAPSGetInfo(ct.byref(si), self.iid,
                                       cyl, head, 1, i)
            error.check(res == 0, "Couldn't get sector info")
            # Adjust the range start to be splice- rather than index-relative
            data.append(((si.datastart - ti.overlap) % ti.tracklen,
                         si.datasize))

        weak = []
        for i in range(ti.weakcnt):
            wi = CapsDataInfo()
            res = self.lib.CAPSGetInfo(ct.byref(wi), self.iid,
                                       cyl, head, 2, i)
            error.check(res == 0, "Couldn't get weak data info")
            # Adjust the range start to be splice- rather than index-relative
            weak.append(((wi.start - ti.overlap) % ti.tracklen, wi.size))

        timebuf = None
        if ti.timebuf:
            carray_type = ct.c_uint * ti.timelen
            carray = carray_type.from_address(
                ct.addressof(ti.timebuf.contents))
            # Unpack the per-byte timing info into per-bitcell
            timebuf = []
            for i in carray:
                for j in range(8):
                    timebuf.append(i)
            # Pad the timing info with normal cell lengths as necessary
            for j in range(len(carray)*8, ti.tracklen):
                timebuf.append(1000)
            # Clip the timing info, if necessary.
            timebuf = timebuf[:ti.tracklen]

        # Rotate the track to start at the splice rather than the index.
        if ti.overlap:
            trackbuf = trackbuf[ti.overlap:] + trackbuf[:ti.overlap]
            if timebuf:
                timebuf = timebuf[ti.overlap:] + timebuf[:ti.overlap]
            
        # We don't really have access to the bitrate. It depends on RPM.
        # So we assume a rotation rate of 300 RPM (5 rev/sec).
        rpm = 300

        track = IPFTrack(
            bits = trackbuf,
            time_per_rev = 60/rpm,
            bit_ticks = timebuf,
            splice = ti.overlap,
            weak = weak)
        track.verify = track
        track.sectors = data
        return track


# Open and initialise the CAPS library.
def open_libcaps():

    # Get the OS-dependent list of valid CAPS library names.
    _names = []
    if platform.system() == "Linux":
        _names = [ "libcapsimage.so.5", "libcapsimage.so.5.1",
                   "libcapsimage.so.4", "libcapsimage.so.4.2",
                   "libcapsimage.so" ]
    elif platform.system() == "Darwin":
        _names = [ "CAPSImage.framework/CAPSImage" ]
    elif platform.system() == "Windows":
        _names = [ "CAPSImg_x64.dll", "CAPSImg.dll" ]

    # Get the absolute path to the root Greaseweazle folder.
    path = os.path.dirname(os.path.abspath(__file__))
    for _ in range(3):
        path = os.path.join(path, os.pardir)
    path = os.path.normpath(path)

    # Create a search list of both relative and absolute library names.
    names = []
    for name in _names:
        names.append(name)
        names.append(os.path.join(path, name))

    # Walk the search list, trying to open the CAPS library.
    for name in names:
        try:
            lib = ct.cdll.LoadLibrary(name)
            break
        except:
            pass
    
    error.check("lib" in locals(), """\
Could not find SPS/CAPS IPF decode library
For installation instructions please read the wiki:
<https://github.com/keirf/greaseweazle/wiki/IPF-Images>""")
    
    # We have opened the library. Now initialise it.
    res = lib.CAPSInit()
    error.check(res == 0, "Failure initialising IPF library '%s'" % name)

    return lib


# Get a reference to the CAPS library. Open it if necessary.
def get_libcaps():
    global libcaps
    if not 'libcaps' in globals():
        libcaps = open_libcaps()
    return libcaps


# Local variables:
# python-indent: 4
# End:



# SPDX-FileCopyrightText: 2022 Jeff Epler for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""Flux visualizer"""

import click
from skimage.io import imsave
import matplotlib.pyplot as plt
from . import open_flux, process


@click.group()
@click.pass_context
@click.option(
    "--slices",
    default=4000,
    help="The angular resolution of the flux in 1/revolution (default: 4000)",
)
@click.option("--stacks", default=6, help="The pixel thickness of a track (default: 6)")
@click.option(
    "--diameter",
    default=216.0,
    help="The diameter of the outermost track, in track-widths."
    'A 5.25" HD disk is approximately 216, 3.5" HD disk is approximately 420. '
    "Halve for DD. (default: 216)",
)
@click.option("--tracks", default=80, help="The total count of tracks (default: 80)")
@click.option("--stride", default=1, help="Stride between tracks (default: 1)")
@click.option("--start", default=0, help="Number of the first tracks (default: 0)")
@click.option("--side", default=0, help="Side of floppy (default: 0)")
@click.option(
    "--resolution", default=960, help="Resolution of square output image (default: 960)"
)
@click.option("--linear/--polar", default=False, help="Image style (default: polar)")
@click.option(
    "--oversample", default=2, help="Increase oversampling of polar transformation"
)
@click.option(
    "--location",
    default=None,
    type=click.Path(exists=True),
    help="fluxengine decoder location information saved with --decoder.write_csv_to=",
)
def main(
    ctx,
    side,
    tracks,
    start,
    stride,
    slices,
    stacks,
    location,
    diameter,
    resolution,
    linear,
    oversample,
):  # pylint: disable=redefined-builtin,too-many-arguments,too-many-locals, invalid-name, too-many-branches,too-many-statements
    """Commandline interface to visualize flux"""
    ctx.ensure_object(dict)
    ctx.obj.update(
        {
            "side": side,
            "tracks": tracks,
            "start": start,
            "stride": stride,
            "linear": linear,
            "slices": slices,
            "stacks": stacks,
            "location": location,
            "diameter": diameter,
            "resolution": resolution,
            "oversample": oversample,
        }
    )


@main.command()
@click.pass_context
@click.argument("input_file", type=click.Path(exists=True))
def view(ctx, input_file):
    """Render a flux image to the screen

    INPUT_FILE may be any flux format recognized by the embedded copy of
    greaseweazle (.scp, etc) or an a2r file."""
    flux = open_flux(input_file)
    density = process(flux, **ctx.obj)

    fig, axis = plt.subplots()
    fig.set_dpi(96)
    fig.set_size_inches(density.shape[0] / 96, density.shape[1] / 96)
    fig.set_frameon(False)
    fig.set_tight_layout(True)
    plt.axis("off")
    axis.imshow(density)
    plt.show()


@main.command()
@click.pass_context
@click.argument("input_file", type=click.Path(exists=True))
@click.argument("output_file", type=click.Path())
def write(ctx, input_file, output_file):
    """Render a flux image to a file

    INPUT_FILE may be any flux format recognized by the embedded copy of
    greaseweazle (.scp, etc) or an a2r file.

    OUTPUT_FILE may be any image format recognized by scikit_img including PNG,
    GIF, and JPG"""
    flux = open_flux(input_file)
    density = process(flux, **ctx.obj)
    imsave(output_file, density)


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter



# greaseweazle/error.py
#
# Error management and reporting.
#
# Written & released by Keir Fraser <keir.xen@gmail.com>
#
# This is free and unencumbered software released into the public domain.
# See the file COPYING for more details, or visit <http://unlicense.org>.

class Fatal(Exception):
    pass

def check(pred, desc):
    if not pred:
        raise Fatal(desc)
    
# Local variables:
# python-indent: 4
# End:



# greaseweazle/image/img.py
#
# Written & released by Keir Fraser <keir.xen@gmail.com>
#
# This is free and unencumbered software released into the public domain.
# See the file COPYING for more details, or visit <http://unlicense.org>.

from .. import error
from ..codec.ibm import mfm
from .image import Image

from ..codec import formats

class IMG(Image):

    sides_swapped = False
    
    def __init__(self, name, fmt):
        self.to_track = dict()
        error.check(fmt is not None and fmt.img_compatible, """\
Sector image requires compatible format conversion
Compatible formats:\n%s"""
                    % formats.print_formats(
                        lambda k, v: v.img_compatible))
        self.filename = name
        self.fmt = fmt


    @classmethod
    def from_file(cls, name, fmt):

        with open(name, "rb") as f:
            dat = f.read()

        img = cls(name, fmt)

        pos = 0
        for t in fmt.max_tracks:
            cyl, head = t.cyl, t.head
            if img.sides_swapped:
                head ^= 1
            track = fmt.fmt(cyl, head)
            pos += track.set_img_track(dat[pos:])
            img.to_track[cyl,head] = track

        return img


    @classmethod
    def to_file(cls, name, fmt, noclobber):
        img = cls(name, fmt)
        img.noclobber = noclobber
        return img

    def get_track(self, cyl, side):
        if (cyl,side) not in self.to_track:
            return None
        return self.to_track[cyl,side].raw_track()


    def emit_track(self, cyl, side, track):
        self.to_track[cyl,side] = track


    def get_image(self):

        tdat = bytearray()

        n_side = 2
        n_cyl = max(self.to_track.keys(), default=(0), key=lambda x:x[0])[0]
        n_cyl += 1

        for cyl in range(n_cyl):
            for head in range(n_side):
                if self.sides_swapped:
                    head ^= 1
                if (cyl,head) in self.to_track:
                    tdat += self.to_track[cyl,head].get_img_track()

        return tdat


# Local variables:
# python-indent: 4
# End:



# SPDX-FileCopyrightText: 2022 Jeff Epler for Adafruit Industries
#
# SPDX-License-Identifier: MIT
"""Visualize floppy flux"""

import csv
import numpy as np
from skimage.color import gray2rgb
from skimage.transform import downscale_local_mean
from .greaseweazle.tools.util import get_image_class
from . import a2rchery
from . import invpolar

PHYSICAL_TRACK = "Physical track"
PHYSICAL_SIDE = "Physical side"
HEADER_START_NS = "Header start (ns)"
HEADER_END_NS = "Header end (ns)"
DATA_START_NS = "Data start (ns)"
DATA_END_NS = "Data end (ns)"
STATUS = "Status"

HEADER_COLOR = np.array((128, 128, 255), dtype=np.uint8)
DATA_OK_COLOR = np.array((128, 255, 128), dtype=np.uint8)
DATA_BAD_COLOR = np.array((255, 128, 128), dtype=np.uint8)


class A2RTrackShim:
    """Adapt an A2R file's track to act similar to a GreaseWeazle track"""

    def __init__(self, flux, est_index):
        self.index_list = [est_index]
        self.list = []
        append = self.list.append
        subtotal = 0
        for value in flux:
            subtotal += value
            if value < 255:
                append(value)
                subtotal = 0

    def cue_at_index(self):
        """No-operation needed for GW compatibility"""


class A2RFluxShim:
    """Adapt an A2R file to act similar to a GreaseWeazle flux file"""

    @property
    def sample_freq(self):
        """The (fixed) sample frequency of an A2R file"""
        return 8_000_000

    def __init__(self, a2r):
        self._a2r = a2r

    def get_track(self, track, side):
        """Retrieve data about a track"""
        if side == 0 and track in self._a2r.flux:
            flux = self._a2r.flux[track][0]
            return A2RTrackShim(flux["data"], flux["tick_count"])
        return A2RTrackShim([1, 1, 1, 1], 4)


def a2r_to_flux(a2r):
    """Wrap a shim around an A2R file so we can visualize it"""
    return A2RFluxShim(a2r)


def open_flux(filename):
    """Open a flux file by filename"""
    if filename.lower().endswith(".a2r"):
        a2r = a2rchery.A2RReader(filename)
        return a2r_to_flux(a2r)
    loader = get_image_class(filename)
    return loader.from_file(filename)


def render_flux(
    flux, side, tracks, start, stride, major, slices, stacks, location
):  # pylint: disable=too-many-arguments,invalid-name,too-many-locals,too-many-branches,too-many-statements
    """Render flux to a linear image"""
    density = np.zeros((major, slices), dtype=np.float32)
    r = np.arange(slices)

    track_duration = {}

    for cyl in range(0, tracks):
        t0 = major - (1 + cyl) * stacks
        t1 = major - cyl * stacks - 1
        track = flux.get_track(cyl * stride + start, side)
        if track is None:
            continue
        track.cue_at_index()
        if not track.index_list:
            index = sum(track.list)
        else:
            index = track.index_list[0]
        index = track.index_list[0]
        track_duration[cyl] = index
        indices = np.array(track.list, dtype=np.float32) * ((slices - 1) / index)
        indices = np.floor(np.add.accumulate(indices))
        lo = np.searchsorted(indices, r, "left")
        hi = np.searchsorted(indices, r, "right")
        density[t0:t1, :] = hi - lo

    def ns2loc(offset_ns, cyl_duration):
        return round(
            offset_ns * flux.sample_freq / 1_000_000_000 / cyl_duration * (slices - 1)
        )

    if location is not None:

        def paint(t0, t1, start, end, color):
            while start > slices:
                start -= slices
                end -= slices
            status[t0 : t1 + 1, start:end] = color
            if end > slices:
                status[t0 : t1 + 1, : end % slices] = color

        status = np.zeros((major, slices, 3), dtype=np.uint8) + 255
        with open(location, "r", encoding="utf-8", errors="ignore") as f:
            for row in csv.DictReader(f):
                physical_track = row.get(PHYSICAL_TRACK)
                physical_side = row.get(PHYSICAL_SIDE)
                header_start = row.get(HEADER_START_NS)
                header_end = row.get(HEADER_END_NS)
                data_start = row.get(DATA_START_NS)
                data_end = row.get(DATA_END_NS)
                data_status = row.get(STATUS)

                if any(x is None for x in (physical_track, physical_side, data_status)):
                    continue

                physical_side = int(physical_side)
                physical_track = int(physical_track)
                if physical_side != side:
                    continue
                if physical_track % stride != start:
                    continue

                cyl = (physical_track - start) // stride
                if cyl not in track_duration:
                    continue
                t0 = major - (1 + cyl) * stacks
                t1 = major - cyl * stacks - 1

                if header_start is not None and header_end is not None:
                    header_start = ns2loc(float(header_start), track_duration[cyl])
                    header_end = ns2loc(float(header_end), track_duration[cyl])
                    if data_start is not None:  # artifically enlarge header mark
                        header_end = ns2loc(float(data_start), track_duration[cyl])
                    paint(t0, t1, header_start, header_end, HEADER_COLOR)

                if data_start is not None and data_end is not None:
                    data_start = ns2loc(float(data_start), track_duration[cyl])
                    data_end = ns2loc(float(data_end), track_duration[cyl])
                    paint(
                        t0,
                        t1,
                        data_start,
                        data_end,
                        DATA_OK_COLOR if data_status == "OK" else DATA_BAD_COLOR,
                    )

        density = gray2rgb(density) * status
    return density


def circularize(density, resolution, oversample):
    """Transform a linear density image to circular"""
    multichannel = len(density.shape) == 3
    density = invpolar.warp_inverse_polar(
        density,
        output_shape=(resolution * oversample, resolution * oversample),
        multichannel=multichannel,
    )
    if multichannel:
        return downscale_local_mean(density, (oversample, oversample, 1))
    return downscale_local_mean(density, (oversample, oversample))


def process(
    flux,
    side,
    tracks,
    start,
    stride,
    linear,
    slices,
    stacks,
    location,
    diameter,
    resolution,
    oversample,
):  # pylint: disable=too-many-arguments
    """Process flux into an image"""

    if linear:
        major = round(tracks * stacks)
    else:
        major = round(diameter * stacks / 2)

    density = render_flux(
        flux, side, tracks, start, stride, major, slices, stacks, location
    )

    if not linear:
        density = circularize(density, resolution, oversample)

    maxdensity = np.max(density)
    return (density * (255 / maxdensity)).astype(np.uint8)



# greaseweazle/image/scp.py
#
# Written & released by Keir Fraser <keir.xen@gmail.com>
#
# This is free and unencumbered software released into the public domain.
# See the file COPYING for more details, or visit <http://unlicense.org>.

import struct, functools

from .. import error
from ..flux import Flux
from .image import Image


# Names for disktype byte in SCP file header
DiskType = {
    'amiga':       0x04,
    'c64':         0x00,
    'atari800-sd': 0x10,
    'atari800-dd': 0x11,
    'atari800-ed': 0x12,
    'atarist-ss':  0x14,
    'atarist-ds':  0x15,
    'appleII':     0x20,
    'appleIIpro':  0x21,
    'apple-400k':  0x24,
    'apple-800k':  0x25,
    'apple-1m44':  0x26,
    'ibmpc-320k':  0x30,
    'ibmpc-720k':  0x31,
    'ibmpc-1m2':   0x32,
    'ibmpc-1m44':  0x33,
    'trs80_sssd':  0x40,
    'trs80_ssdd':  0x41,
    'trs80_dssd':  0x42,
    'trs80_dsdd':  0x43,
    'ti-99/4a':    0x50,
    'roland-d20':  0x60,
    'amstrad-cpc': 0x70,
    'other-320k':  0x80,
    'other-1m2':   0x81,
    'other-720k':  0x84,
    'other-1m44':  0x85,
    'tape-gcr1':   0xe0,
    'tape-gcr2':   0xe1,
    'tape-mfm':    0xe2,
    'hdd-mfm':     0xf0,
    'hdd-rll':     0xf1
}


class SCPOpts:
    """legacy_ss: Set to True to generate (incorrect) legacy single-sided
    SCP image.
    """

    def __init__(self):
        self.legacy_ss = False
        self._disktype = 0x80 # Other

    @property
    def disktype(self):
        return self._disktype
    @disktype.setter
    def disktype(self, disktype):
        try:
            self._disktype = DiskType[disktype.lower()]
        except KeyError:
            try:
                self._disktype = int(disktype, 0)
            except ValueError:
                raise error.Fatal("Bad SCP disktype: '%s'" % disktype)


class SCPTrack:

    def __init__(self, tdh, dat, splice=None):
        self.tdh = tdh
        self.dat = dat
        self.splice = splice


class SCP(Image):

    # 40MHz
    sample_freq = 40000000


    def __init__(self):
        self.opts = SCPOpts()
        self.nr_revs = None
        self.to_track = dict()
        self.index_cued = True

    
    def side_count(self):
        s = [0,0] # non-empty tracks on each side
        for tnr in self.to_track:
            s[tnr&1] += 1
        return s


    @classmethod
    def from_file(cls, name):

        splices = None

        with open(name, "rb") as f:
            dat = f.read()

        header = struct.unpack("<3s9BI", dat[0:16])
        sig, _, disk_type, nr_revs, _, _, flags, _, single_sided, _, _ = header
        error.check(sig == b"SCP", "SCP: Bad signature")

        index_cued = flags & 1 or nr_revs == 1

        # Some tools generate a short TLUT. We handle this by truncating the
        # TLUT at the first Track Data Header.
        trk_offs = struct.unpack("<168I", dat[16:0x2b0])
        for i in range(168):
            try:
                off = trk_offs[i]
            except IndexError:
                break
            if off == 0 or off >= 0x2b0:
                continue
            off = off//4 - 4
            error.check(off >= 0, "SCP: Bad Track Table")
            trk_offs = trk_offs[:off]

        # Parse the extension block introduced by github:markusC64/g64conv.
        # b'EXTS', length, <length bytes: Extension Area>
        # Extension Area contains consecutive chunks of the form:
        # ID, length, <length bytes: ID-specific data>
        ext_sig, ext_len = struct.unpack('<4sI', dat[0x2b0:0x2b8])
        min_tdh = min(filter(lambda x: x != 0, trk_offs), default=0)
        if ext_sig == b'EXTS' and 0x2b8 + ext_len <= min_tdh:
            pos, end = 0x2b8, 0x2b8 + ext_len
            while end - pos >= 8:
                chk_sig, chk_len = struct.unpack('<4sI', dat[pos:pos+8])
                pos += 8
                # WRSP: WRite SPlice information block.
                # Data is comprised of >= 169 32-bit values:
                #  0: Flags (currently unused; must be zero)
                #  N: Write splice/overlap position for track N, in SCP ticks
                #     (zero if the track is unused)
                if chk_sig == b'WRSP' and chk_len >= 169*4:
                    # Write-splice positions for writing out SCP tracks
                    # correctly to disk.
                    splices = struct.unpack('<168I', dat[pos+4:pos+169*4])
                pos += chk_len

        scp = cls()

        for trknr in range(len(trk_offs)):
            
            trk_off = trk_offs[trknr]
            if trk_off == 0:
                continue

            # Parse the SCP track header and extract the flux data.
            thdr = dat[trk_off:trk_off+4+12*nr_revs]
            sig, tnr = struct.unpack("<3sB", thdr[:4])
            error.check(sig == b"TRK", "SCP: Missing track signature")
            error.check(tnr == trknr, "SCP: Wrong track number in header")
            thdr = thdr[4:] # Remove TRK header

            # Strip empty trailing revolutions (old versions of FluxEngine).
            while thdr:
                e_ticks, e_nr, e_off = struct.unpack("<3I", thdr[-12:])
                if e_nr != 0 and e_ticks != 0:
                    break
                thdr = thdr[:-12]
            # Bail if all revolutions are empty.
            if not thdr:
                continue

            # Clip the first revolution if it's not flagged as index cued
            # and there's more than one revolution.
            if not index_cued and len(thdr) > 12:
                thdr = thdr[12:]
            s_off, = struct.unpack("<I", thdr[8:12])

            e_off += e_nr*2
            if s_off == e_off:
                # FluxEngine creates dummy TDHs for empty tracks.
                # Bail on them here.
                continue

            tdat = dat[trk_off+s_off:trk_off+e_off]
            track = SCPTrack(thdr, tdat)
            if splices is not None:
                track.splice = splices[trknr]
            scp.to_track[trknr] = track

        s = scp.side_count()

        # C64 images with halftracks are genberated by Supercard Pro using
        # consecutive track numbers. That needs fixup here for our layout.
        # We re-use the legacy-single-sided fixup below.
        if (single_sided == 0 and disk_type == 0
            and s[1] and s[0]==s[1]+1 and s[0] < 42):
            single_sided = 1
            print('SCP: Importing C64 image with halftracks')

        # Some tools produce (or used to produce) single-sided images using
        # consecutive entries in the TLUT. This needs fixing up.
        if single_sided and s[0] and s[1]:
            new_dict = dict()
            for tnr in scp.to_track:
                new_dict[tnr*2+single_sided-1] = scp.to_track[tnr]
            scp.to_track = new_dict
            print('SCP: Imported legacy single-sided image')

        return scp


    def get_track(self, cyl, side):
        tracknr = cyl * 2 + side
        if not tracknr in self.to_track:
            return None
        track = self.to_track[tracknr]
        tdh, dat = track.tdh, track.dat

        index_list = []
        while tdh:
            ticks, _, _ = struct.unpack("<3I", tdh[:12])
            index_list.append(ticks)
            tdh = tdh[12:]
        
        # Decode the SCP flux data into a simple list of flux times.
        flux_list = []
        val = 0
        for i in range(0, len(dat), 2):
            x = dat[i]*256 + dat[i+1]
            if x == 0:
                val += 65536
                continue
            flux_list.append(val + x)
            val = 0

        flux = Flux(index_list, flux_list, SCP.sample_freq)
        flux.splice = track.splice if track.splice is not None else 0
        return flux


    def emit_track(self, cyl, side, track):
        """Converts @track into a Supercard Pro Track and appends it to
        the current image-in-progress.
        """

        flux = track.flux()

        # External tools and emulators generally seem to work best (or only)
        # with index-cued SCP image files. So let's make sure we give them
        # what they want.
        flux.cue_at_index()

        if not flux.index_cued:
            self.index_cued = False

        nr_revs = len(flux.index_list)
        if not self.nr_revs:
            self.nr_revs = nr_revs
        else:
            assert self.nr_revs == nr_revs
        
        factor = SCP.sample_freq / flux.sample_freq

        tdh, dat = bytearray(), bytearray()
        len_at_index = rev = 0
        to_index = flux.index_list[0]
        rem = 0.0

        for x in flux.list:

            # Does the next flux interval cross the index mark?
            while to_index < x:
                # Append to the TDH for the previous full revolution
                tdh += struct.pack("<III",
                                   round(flux.index_list[rev]*factor),
                                   (len(dat) - len_at_index) // 2,
                                   4 + nr_revs*12 + len_at_index)
                # Set up for the next revolution
                len_at_index = len(dat)
                rev += 1
                if rev >= nr_revs:
                    # We're done: We simply discard any surplus flux samples
                    self.to_track[cyl*2+side] = SCPTrack(tdh, dat)
                    return
                to_index += flux.index_list[rev]

            # Process the current flux sample into SCP "bitcell" format
            to_index -= x
            y = x * factor + rem
            val = round(y)
            if (val & 65535) == 0:
                val += 1
            rem = y - val
            while val >= 65536:
                dat.append(0)
                dat.append(0)
                val -= 65536
            dat.append(val>>8)
            dat.append(val&255)

        # Header for last track(s) in case we ran out of flux timings.
        while rev < nr_revs:
            tdh += struct.pack("<III",
                               round(flux.index_list[rev]*factor),
                               (len(dat) - len_at_index) // 2,
                               4 + nr_revs*12 + len_at_index)
            len_at_index = len(dat)
            rev += 1

        self.to_track[cyl*2+side] = SCPTrack(tdh, dat)


    def get_image(self):

        # Work out the single-sided byte code
        s = self.side_count()
        if s[0] and s[1]:
            single_sided = 0
        elif s[0]:
            single_sided = 1
        else:
            single_sided = 2

        to_track = self.to_track
        if single_sided and self.opts.legacy_ss:
            print('SCP: Generated legacy single-sided image')
            to_track = dict()
            for tnr in self.to_track:
                to_track[tnr//2] = self.to_track[tnr]

        ntracks = max(to_track, default=0) + 1

        # Generate the TLUT and concatenate all the tracks together.
        trk_offs = bytearray()
        trk_dat = bytearray()
        for tnr in range(ntracks):
            if tnr in to_track:
                track = to_track[tnr]
                trk_offs += struct.pack("<I", 0x2b0 + len(trk_dat))
                trk_dat += struct.pack("<3sB", b"TRK", tnr)
                trk_dat += track.tdh + track.dat
            else:
                trk_offs += struct.pack("<I", 0)
        error.check(len(trk_offs) <= 0x2a0, "SCP: Too many tracks")
        trk_offs += bytes(0x2a0 - len(trk_offs))

        # Calculate checksum over all data (except 16-byte image header).
        csum = 0
        for x in trk_offs:
            csum += x
        for x in trk_dat:
            csum += x

        # Generate the image header.
        flags = 2 # 96TPI
        if self.index_cued:
            flags |= 1 # Index-Cued
        header = struct.pack("<3s9BI",
                             b"SCP",    # Signature
                             0,         # Version
                             self.opts.disktype,
                             self.nr_revs, 0, ntracks-1,
                             flags,
                             0,         # 16-bit cell width
                             single_sided,
                             0,         # 25ns capture
                             csum & 0xffffffff)

        # Concatenate it all together and send it back.
        return header + trk_offs + trk_dat


# Local variables:
# python-indent: 4
# End:



# greaseweazle/image/edsk.py
#
# Some of the code here is heavily inspired by Simon Owen's SAMdisk:
# https://simonowen.com/samdisk/
#
# Written & released by Keir Fraser <keir.xen@gmail.com>
# 
# This is free and unencumbered software released into the public domain.
# See the file COPYING for more details, or visit <http://unlicense.org>.

import binascii, math, struct
import itertools as it
from bitarray import bitarray

from .. import error
from ..codec.ibm import mfm
from ..track import MasterTrack, RawTrack
from .image import Image

class SR1:
    SUCCESS                   = 0x00
    CANNOT_FIND_ID_ADDRESS    = 0x01
    WRITE_PROTECT_DETECTED    = 0x02
    CANNOT_FIND_SECTOR_ID     = 0x04
    RESERVED1                 = 0x08
    OVERRUN                   = 0x10
    CRC_ERROR                 = 0x20
    RESERVED2                 = 0x40
    END_OF_CYLINDER           = 0x80

class SR2:
    SUCCESS                   = 0x00
    MISSING_ADDRESS_MARK      = 0x01
    BAD_CYLINDER              = 0x02
    SCAN_COMMAND_FAILED       = 0x04
    SCAN_COMMAND_EQUAL        = 0x08
    WRONG_CYLINDER_DETECTED   = 0x10
    CRC_ERROR_IN_SECTOR_DATA  = 0x20
    SECTOR_WITH_DELETED_DATA  = 0x40
    RESERVED                  = 0x80

class SectorErrors:
    def __init__(self, sr1, sr2):
        self.id_crc_error = (sr1 & SR1.CRC_ERROR) != 0
        self.data_not_found = (sr2 & SR2.MISSING_ADDRESS_MARK) != 0
        self.data_crc_error = (sr2 & SR2.CRC_ERROR_IN_SECTOR_DATA) != 0
        self.deleted_dam = (sr2 & SR2.SECTOR_WITH_DELETED_DATA) != 0
        if self.data_crc_error:
            # uPD765 sets both id and data flags for data CRC errors
            self.id_crc_error = False
        if (# normal data
            (sr1 == SR1.SUCCESS and sr2 == SR2.SUCCESS) or
            # deleted data
            (sr1 == SR1.SUCCESS and sr2 == SR2.SECTOR_WITH_DELETED_DATA) or
            # end of track
            (sr1 == SR1.END_OF_CYLINDER and sr2 == SR2.SUCCESS) or
            # id crc error
            (sr1 == SR1.CRC_ERROR and sr2 == SR2.SUCCESS) or
            # normal data crc error
            (sr1 == SR1.CRC_ERROR and sr2 == SR2.CRC_ERROR_IN_SECTOR_DATA) or
            # deleted data crc error
            (sr1 == SR1.CRC_ERROR and sr2 == (SR2.CRC_ERROR_IN_SECTOR_DATA |
                                              SR2.SECTOR_WITH_DELETED_DATA)) or
            # data field missing (some FDCs set AM in ST1)
            (sr1 == SR1.CANNOT_FIND_ID_ADDRESS
             and sr2 == SR2.MISSING_ADDRESS_MARK) or
            # data field missing (some FDCs don't)
            (sr1 == SR1.SUCCESS and sr2 == SR2.MISSING_ADDRESS_MARK) or
            # CHRN mismatch
            (sr1 == SR1.CANNOT_FIND_SECTOR_ID and sr2 == SR2.SUCCESS) or
            # CHRN mismatch, including wrong cylinder
            (sr1 == SR1.CANNOT_FIND_SECTOR_ID
             and sr2 == SR2.WRONG_CYLINDER_DETECTED)):
            pass
        else:
            print('Unusual status flags (ST1=%02X ST2=%02X)' % (sr1, sr2))
            
class EDSKTrack:

    gap_presync = 12
    gap_4a = 80 # Post-Index
    gap_1  = 50 # Post-IAM
    gap_2  = 22 # Post-IDAM

    gapbyte = 0x4e
    
    def __init__(self):
        self.time_per_rev = 0.2
        self.clock = 2e-6
        self.bits, self.weak, self.bytes = [], [], bytearray()

    def raw_track(self):
        track = MasterTrack(
            bits = self.bits,
            time_per_rev = self.time_per_rev,
            weak = self.weak)
        track.verify = self
        track.verify_revs = 1
        return track

    def _find_sync(self, bits, sync, start):
        for offs in bits.itersearch(sync):
            if offs >= start:
                return offs
        return None
    
    def verify_track(self, flux):
        flux.cue_at_index()
        raw = RawTrack(time_per_rev = self.time_per_rev,
                       clock = self.time_per_rev / len(self.bits),
                       data = flux)
        bits, _ = raw.get_all_data()
        weak_iter = it.chain(self.weak, [(self.verify_len+1,1)])
        weak = next(weak_iter)

        # Start checking from the IAM sync
        dump_start = self._find_sync(bits, mfm.iam_sync, 0)
        self_start = self._find_sync(self.bits, mfm.iam_sync, 0)

        # Include the IAM pre-sync header
        if dump_start is None:
            return False
        dump_start -= self.gap_presync * 16
        self_start -= self.gap_presync * 16

        while self_start is not None and dump_start is not None:

            # Find the weak areas immediately before and after the current
            # region to be checked.
            s,n = None,None
            while self_start > weak[0]:
                s,n = weak
                weak = next(weak_iter)

            # If there is a weak area preceding us, move the start point to
            # immediately follow the weak area.
            if s is not None:
                delta = self_start - (s + n + 16)
                self_start -= delta
                dump_start -= delta

            # Truncate the region at the next weak area, or the last sector.
            self_end = max(self_start, min(weak[0], self.verify_len+1))
            dump_end = dump_start + self_end - self_start

            # Extract the corresponding areas from the pristine track and
            # from the dump, and check that they match.
            if bits[dump_start:dump_end] != self.bits[self_start:self_end]:
                return False

            # Find the next A1A1A1 sync pattern
            dump_start = self._find_sync(bits, mfm.sync, dump_end)
            self_start = self._find_sync(self.bits, mfm.sync, self_end)

        # Did we verify all regions in the pristine track?
        return self_start is None

class EDSK(Image):

    read_only = True

    def __init__(self):
        self.to_track = dict()

    # Find all weak ranges in the given sector data copies.
    @staticmethod
    def find_weak_ranges(dat, size):
        orig = dat[:size]
        s, w = size, []
        # Find first mismatching byte across all copies
        for i in range(1, len(dat)//size):
            diff = [x^y for x, y in zip(orig, dat[size*i:size*(i+1)])]
            weak = [idx for idx, val in enumerate(diff) if val != 0]
            if weak:
                s = min(s, weak[0])
        # Look for runs of filler
        i = s
        while i < size:
            j, x = i, orig[i]
            while j < size and orig[j] == x:
                j += 1
            if j-i >= 16:
                w.append((s,i-s))
                s = j
            i = j
        # Append final weak area if any.
        if s < size:
            w.append((s,size-s))
        return w

    @staticmethod
    def _build_8k_track(sectors):
        if len(sectors) != 1:
            return None
        c,h,r,n,errs,data = sectors[0]
        if n != 6:
            return None
        if errs.id_crc_error or errs.data_not_found or not errs.data_crc_error:
            return None
        # Magic longtrack value is for Coin-Op Hits. Taken from SAMdisk.
        if len(data) > 6307:
            data = data[:6307]
        track = EDSKTrack()
        t = track.bytes
        # Post-index gap
        t += mfm.encode(bytes([track.gapbyte] * 16))
        # IAM
        t += mfm.encode(bytes(track.gap_presync))
        t += mfm.iam_sync_bytes
        t += mfm.encode(bytes([mfm.IBM_MFM.IAM]))
        t += mfm.encode(bytes([track.gapbyte] * 16))
        # IDAM
        t += mfm.encode(bytes(track.gap_presync))
        t += mfm.sync_bytes
        am = bytes([0xa1, 0xa1, 0xa1, mfm.IBM_MFM.IDAM, c, h, r, n])
        crc = mfm.crc16.new(am).crcValue
        am += struct.pack('>H', crc)
        t += mfm.encode(am[3:])
        t += mfm.encode(bytes([track.gapbyte] * track.gap_2))
        # DAM
        t += mfm.encode(bytes(track.gap_presync))
        t += mfm.sync_bytes
        dmark = (mfm.IBM_MFM.DDAM if errs.deleted_dam
                 else mfm.IBM_MFM.DAM)
        am = bytes([0xa1, 0xa1, 0xa1, dmark]) + data
        t += mfm.encode(am[3:])
        return track

    @staticmethod
    def _build_kbi19_track(sectors):
        ids = [0,1,4,7,10,13,16,2,5,8,11,14,17,3,6,9,12,15,18]
        if len(sectors) != len(ids):
            return None
        for s,id in zip(sectors,ids):
            c,h,r,n,_,_ = s
            if r != id or n != 2:
                return None
        def addcrc(t,n):
            crc = mfm.crc16.new(mfm.decode(t[-n*2:])).crcValue
            t += mfm.encode(struct.pack('>H', crc))
        track = EDSKTrack()
        t = track.bytes
        # Post-index gap
        t += mfm.encode(bytes([track.gapbyte] * 64))
        # IAM
        t += mfm.encode(bytes(track.gap_presync))
        t += mfm.iam_sync_bytes
        t += mfm.encode(bytes([mfm.IBM_MFM.IAM]))
        t += mfm.encode(bytes([track.gapbyte] * 50))
        for idx, s in enumerate(sectors):
            c,h,r,n,errs,data = s
            # IDAM
            t += mfm.encode(bytes(track.gap_presync))
            t += mfm.sync_bytes
            t += mfm.encode(bytes([mfm.IBM_MFM.IDAM, c, h, r, n]))
            addcrc(t, 8)
            if r == 0:
                t += mfm.encode(bytes([track.gapbyte] * 17))
                t += mfm.encode(b' KBI ')
            else:
                t += mfm.encode(bytes([track.gapbyte] * 8))
                t += mfm.encode(b' KBI ')
                t += mfm.encode(bytes([track.gapbyte] * 9))
            # DAM
            t += mfm.encode(bytes(track.gap_presync))
            t += mfm.sync_bytes
            dmark = (mfm.IBM_MFM.DDAM if errs.deleted_dam
                     else mfm.IBM_MFM.DAM)
            t += mfm.encode(bytes([dmark]))
            if idx%3 != 0:
                t += mfm.encode(data[:61])
            elif r == 0:
                t += mfm.encode(data[:512])
                addcrc(t,516)
            else:
                t += mfm.encode(data[0:0x10e])
                addcrc(t,516)
                t += mfm.encode(data[0x110:0x187])
                addcrc(t,516)
                t += mfm.encode(data[0x189:0x200])
                addcrc(t,516)
                t += mfm.encode(bytes([track.gapbyte] * 80))
        return track

    @classmethod
    def from_file(cls, name):

        with open(name, "rb") as f:
            dat = f.read()

        edsk = cls()

        sig, creator, ncyls, nsides, track_sz = struct.unpack(
            '<34s14s2BH', dat[:52])
        if sig[:8] == b'MV - CPC':
            extended = False
        elif sig[:16] == b'EXTENDED CPC DSK':
            extended = True
        else:
            raise error.Fatal('Unrecognised CPC DSK file: bad signature')

        if extended:
            track_sizes = list(dat[52:52+ncyls*nsides])
            track_sizes = list(map(lambda x: x*256, track_sizes))
        else:
            track_sizes = [track_sz] * (ncyls * nsides)

        o = 256 # skip disk header and track-size table
        for track_size in track_sizes:
            if track_size == 0:
                continue
            sig, cyl, head, sec_sz, nsecs, gap_3, filler = struct.unpack(
                '<12s4x2B2x4B', dat[o:o+24])
            error.check(sig == b'Track-Info\r\n',
                        'EDSK: Missing track header')
            error.check((cyl, head) not in edsk.to_track,
                        'EDSK: Track specified twice')
            bad_crc_clip_data = False
            while True:
                track = EDSKTrack()
                t = track.bytes
                # Post-index gap
                t += mfm.encode(bytes([track.gapbyte] * track.gap_4a))
                # IAM
                t += mfm.encode(bytes(track.gap_presync))
                t += mfm.iam_sync_bytes
                t += mfm.encode(bytes([mfm.IBM_MFM.IAM]))
                t += mfm.encode(bytes([track.gapbyte] * track.gap_1))
                sh = dat[o+24:o+24+8*nsecs]
                data_pos = o + 256 # skip track header and sector-info table
                clippable, ngap3, sectors, idam_included = 0, 0, [], False
                while sh:
                    c, h, r, n, stat1, stat2, data_size = struct.unpack(
                        '<6BH', sh[:8])
                    sh = sh[8:]
                    native_size = mfm.sec_sz(n)
                    weak = []
                    errs = SectorErrors(stat1, stat2)
                    num_copies = 0 if errs.data_not_found else 1
                    if not extended:
                        data_size = mfm.sec_sz(sec_sz)
                    sec_data = dat[data_pos:data_pos+data_size]
                    data_pos += data_size
                    if (extended
                        and data_size > native_size
                        and errs.data_crc_error
                        and (data_size % native_size == 0
                             or data_size == 49152)):
                        num_copies = (3 if data_size == 49152
                                      else data_size // native_size)
                        data_size //= num_copies
                        weak = cls().find_weak_ranges(sec_data, data_size)
                        sec_data = sec_data[:data_size]
                    sectors.append((c,h,r,n,errs,sec_data))
                    # IDAM
                    if not idam_included:
                        t += mfm.encode(bytes(track.gap_presync))
                        t += mfm.sync_bytes
                        am = bytes([0xa1, 0xa1, 0xa1, mfm.IBM_MFM.IDAM,
                                    c, h, r, n])
                        crc = mfm.crc16.new(am).crcValue
                        if errs.id_crc_error:
                            crc ^= 0x5555
                        am += struct.pack('>H', crc)
                        t += mfm.encode(am[3:])
                        t += mfm.encode(bytes([track.gapbyte] * track.gap_2))
                    # DAM
                    gap_included, idam_included = False, False
                    if errs.id_crc_error or errs.data_not_found:
                        continue
                    t += mfm.encode(bytes(track.gap_presync))
                    t += mfm.sync_bytes
                    track.weak += [((s+len(t)//2+1)*16, n*16) for s,n in weak]
                    dmark = (mfm.IBM_MFM.DDAM if errs.deleted_dam
                             else mfm.IBM_MFM.DAM)
                    if errs.data_crc_error:
                        if sh:
                            # Look for next IDAM
                            idam = bytes([0]*12 + [0xa1]*3
                                         + [mfm.IBM_MFM.IDAM])
                            idx = sec_data.find(idam)
                        else:
                            # Last sector: Look for GAP3
                            idx = sec_data.find(bytes([track.gapbyte]*8))
                        if idx > 0:
                            # 2 + gap_3 = CRC + GAP3 (because gap_included)
                            clippable += data_size - idx + 2 + gap_3
                            if bad_crc_clip_data:
                                data_size = idx
                                sec_data = sec_data[:data_size]
                                gap_included = True
                    elif data_size < native_size:
                        # Pad short data
                        sec_data += bytes(native_size - data_size)
                    elif data_size > native_size:
                        # Clip long data if it includes pre-sync 00 bytes
                        if (sec_data[-13] != 0
                            and all([v==0 for v in sec_data[-12:]])):
                            # Includes next pre-sync: Clip it.
                            sec_data = sec_data[:-12]
                        if sh:
                            # Look for next IDAM
                            idam = bytes([0]*12 + [0xa1]*3 + [mfm.IBM_MFM.IDAM]
                                         + list(sh[:4]))
                            idx = sec_data.find(idam)
                            if idx > native_size:
                                # Sector data includes next IDAM. Output it
                                # here and skip it on next iteration.
                                t += mfm.encode(bytes([dmark]))
                                t += mfm.encode(sec_data[:idx+12])
                                t += mfm.sync_bytes
                                t += mfm.encode(sec_data[idx+12+3:])
                                idam_included = True
                                continue
                        # Long data includes CRC and GAP
                        gap_included = True
                    if gap_included:
                        t += mfm.encode(bytes([dmark]))
                        t += mfm.encode(sec_data)
                        continue
                    am = bytes([0xa1, 0xa1, 0xa1, dmark]) + sec_data
                    crc = mfm.crc16.new(am).crcValue
                    if errs.data_crc_error:
                        crc ^= 0x5555
                    am += struct.pack('>H', crc)
                    t += mfm.encode(am[3:])
                    if sh:
                        # GAP3 for all but last sector
                        t += mfm.encode(bytes([track.gapbyte] * gap_3))
                        ngap3 += 1

                # Special track handlers
                special_track = cls()._build_8k_track(sectors)
                if special_track is None:
                    special_track = cls()._build_kbi19_track(sectors)
                if special_track is not None:
                    track = special_track
                    break

                # The track may be too long to fit: Check for overhang.
                tracklen = int((track.time_per_rev / track.clock) / 16)
                overhang = int(len(t)//2 - tracklen*0.99)
                if overhang <= 0:
                    break

                # Some EDSK tracks with Bad CRC contain a raw dump following
                # the DAM. This can usually be clipped.
                if clippable and not bad_crc_clip_data:
                    bad_crc_clip_data = True
                    continue

                # Some EDSK images have bogus GAP3 values. Shrink it if
                # necessary.
                new_gap_3 = -1
                if ngap3 != 0:
                    new_gap_3 = gap_3 - math.ceil(overhang / ngap3)
                error.check(new_gap_3 >= 0,
                            'EDSK: Track %d.%d is too long '
                            '(%d bits @ GAP3=%d; %d bits @ GAP3=0)'
                            % (cyl, head, len(t)*8, gap_3,
                               (len(t)//2-gap_3*ngap3)*16))
                #print('EDSK: GAP3 reduced (%d -> %d)' % (gap_3, new_gap_3))
                gap_3 = new_gap_3

            # Pre-index gap
            track.verify_len = len(track.bytes)*8
            tracklen = int((track.time_per_rev / track.clock) / 16)
            gap = max(40, tracklen - len(t)//2)
            track.bytes += mfm.encode(bytes([track.gapbyte] * gap))

            # Add the clock buts
            track.bits = bitarray(endian='big')
            track.bits.frombytes(mfm.mfm_encode(track.bytes))

            # Register the track
            edsk.to_track[cyl,head] = track
            o += track_size

        return edsk


    def get_track(self, cyl, side):
        if (cyl,side) not in self.to_track:
            return None
        return self.to_track[cyl,side].raw_track()


# Local variables:
# python-indent: 4
# End:



# greaseweazle/image/hfe.py
#
# Written & released by Keir Fraser <keir.xen@gmail.com>
#
# This is free and unencumbered software released into the public domain.
# See the file COPYING for more details, or visit <http://unlicense.org>.

import struct

from .. import error
from ..track import MasterTrack, RawTrack
from bitarray import bitarray
from .image import Image

class HFEOpts:
    """bitrate: Bitrate of new HFE image file.
    """
    
    def __init__(self):
        self._bitrate = None

    @property
    def bitrate(self):
        return self._bitrate
    @bitrate.setter
    def bitrate(self, bitrate):
        try:
            self._bitrate = int(bitrate)
            if self._bitrate <= 0:
                raise ValueError
        except ValueError:
            raise error.Fatal("HFE: Invalid bitrate: '%s'" % bitrate)


class HFE(Image):

    def __init__(self):
        self.opts = HFEOpts()
        # Each track is (bitlen, rawbytes).
        # rawbytes is a bytes() object in little-endian bit order.
        self.to_track = dict()


    @classmethod
    def from_file(cls, name):

        with open(name, "rb") as f:
            dat = f.read()

        (sig, f_rev, n_cyl, n_side, t_enc, bitrate,
         _, _, _, tlut_base) = struct.unpack("<8s4B2H2BH", dat[:20])
        error.check(sig != b"HXCHFEV3", "HFEv3 is not supported")
        error.check(sig == b"HXCPICFE" and f_rev <= 1, "Not a valid HFE file")
        error.check(0 < n_cyl, "HFE: Invalid #cyls")
        error.check(0 < n_side < 3, "HFE: Invalid #sides")

        hfe = cls()
        hfe.opts.bitrate = bitrate

        tlut = dat[tlut_base*512:tlut_base*512+n_cyl*4]
        
        for cyl in range(n_cyl):
            for side in range(n_side):
                offset, length = struct.unpack("<2H", tlut[cyl*4:(cyl+1)*4])
                todo = length // 2
                tdat = bytes()
                while todo:
                    d_off = offset*512 + side*256
                    d_nr = 256 if todo > 256 else todo
                    tdat += dat[d_off:d_off+d_nr]
                    todo -= d_nr
                    offset += 1
                hfe.to_track[cyl,side] = (len(tdat)*8, tdat)

        return hfe


    def get_track(self, cyl, side):
        if (cyl,side) not in self.to_track:
            return None
        bitlen, rawbytes = self.to_track[cyl,side]
        tdat = bitarray(endian='little')
        tdat.frombytes(rawbytes)
        track = MasterTrack(
            bits = tdat[:bitlen],
            time_per_rev = bitlen / (2000*self.opts.bitrate))
        return track


    def emit_track(self, cyl, side, track):
        if self.opts.bitrate is None:
            t = track.raw_track() if hasattr(track, 'raw_track') else track
            b = getattr(t, 'bitrate', None)
            error.check(hasattr(t, 'bitrate'),
                        'HFE: Requires bitrate to be specified'
                        ' (eg. filename.hfe::bitrate=500)')
            self.opts.bitrate = round(t.bitrate / 2e3)
            print('HFE: Data bitrate detected: %d kbit/s' % self.opts.bitrate)
        flux = track.flux()
        flux.cue_at_index()
        raw = RawTrack(clock = 5e-4 / self.opts.bitrate, data = flux)
        bits, _ = raw.get_revolution(0)
        bits.bytereverse()
        self.to_track[cyl,side] = (len(bits), bits.tobytes())


    def get_image(self):

        n_side = 1
        n_cyl = max(self.to_track.keys(), default=(0), key=lambda x:x[0])[0]
        n_cyl += 1

        # We dynamically build the Track-LUT and -Data arrays.
        tlut = bytearray()
        tdat = bytearray()

        # Stuff real data into the image.
        for i in range(n_cyl):
            s0 = self.to_track[i,0] if (i,0) in self.to_track else None
            s1 = self.to_track[i,1] if (i,1) in self.to_track else None
            if s0 is None and s1 is None:
                # Dummy data for empty cylinders. Assumes 300RPM.
                nr_bytes = 100 * self.opts.bitrate
                tlut += struct.pack("<2H", len(tdat)//512 + 2, nr_bytes)
                tdat += bytes([0x88] * (nr_bytes+0x1ff & ~0x1ff))
            else:
                # At least one side of this cylinder is populated.
                if s1 is not None:
                    n_side = 2
                bc = [s0 if s0 is not None else (0,bytes()),
                      s1 if s1 is not None else (0,bytes())]
                nr_bytes = max(len(t[1]) for t in bc)
                nr_blocks = (nr_bytes + 0xff) // 0x100
                tlut += struct.pack("<2H", len(tdat)//512 + 2, 2 * nr_bytes)
                for b in range(nr_blocks):
                    for t in bc:
                        slice = t[1][b*256:(b+1)*256]
                        tdat += slice + bytes([0x88] * (256 - len(slice)))

        # Construct the image header.
        header = struct.pack("<8s4B2H2BH",
                             b"HXCPICFE",
                             0,
                             n_cyl,
                             n_side,
                             0xff, # unknown encoding
                             self.opts.bitrate,
                             0,    # rpm (unused)
                             0xff, # unknown interface
                             1,    # rsvd
                             1)    # track list offset

        # Pad the header and TLUT to 512-byte blocks.
        header += bytes([0xff] * (0x200 - len(header)))
        tlut += bytes([0xff] * (0x200 - len(tlut)))

        return header + tlut + tdat


# Local variables:
# python-indent: 4
# End:



# SPDX-FileCopyrightText: 2022 Jeff Epler for Adafruit Industries
#
# SPDX-License-Identifier: MIT

"""Helper function for use in Jupyter Notebook"""

import io
import os
import tempfile
import shutil
from ipywidgets import FileUpload
from IPython.display import display
import matplotlib.pyplot as plt
from . import open_flux, process


def go(
    side=0,
    tracks=35,
    start=0,
    stride=4,
    linear=False,
    slices=800,
    stacks=3,
    location=None,
    diameter=108,
    resolution=900,
    oversample=2,
):  # pylint: disable=invalid-name,too-many-arguments
    """Main helper function for operation in a notebook"""
    uploader = FileUpload(accept=".a2r,.scp", multiple=False)

    display(uploader)

    def on_upload_change(_):
        try:
            for meta, content in zip(uploader.metadata, uploader.data):
                print(f"processing {meta['name']} with content of {len(content)} bytes")
                process_one_flux(meta["name"], content)
        finally:
            uploader.metadata.clear()
            uploader.data.clear()
            uploader._counter = 0  # pylint: disable=protected-access

    def process_one_flux(filename, content):
        with io.BytesIO(content) as b, tempfile.NamedTemporaryFile(
            suffix=os.path.splitext(filename)[1]
        ) as t:
            shutil.copyfileobj(b, t)
            t.flush()
            flux = open_flux(t.name)
        density = process(
            flux,
            side=side,
            tracks=tracks,
            start=start,
            stride=stride,
            linear=linear,
            slices=slices,
            stacks=stacks,
            location=location,
            diameter=diameter,
            resolution=resolution,
            oversample=oversample,
        )

        fig, axis = plt.subplots()
        fig.set_dpi(96)
        fig.set_size_inches(density.shape[0] / 96, density.shape[1] / 96)
        fig.set_frameon(False)
        fig.set_tight_layout(True)
        plt.axis("off")
        axis.imshow(density)
        plt.show()

    uploader.observe(on_upload_change, names="_counter")



# greaseweazle/flux.py
#
# Written & released by Keir Fraser <keir.xen@gmail.com>
#
# This is free and unencumbered software released into the public domain.
# See the file COPYING for more details, or visit <http://unlicense.org>.

from . import error


class Flux:

    def __init__(self, index_list, flux_list, sample_freq, index_cued=True):
        self.index_list = index_list
        self.list = flux_list
        self.sample_freq = sample_freq
        self.splice = 0
        self.index_cued = index_cued


    def __str__(self):
        s = "\nFlux: %.2f MHz" % (self.sample_freq*1e-6)
        s += ("\n Total: %u samples, %.2fms\n"
              % (len(self.list), sum(self.list)*1000/self.sample_freq))
        rev = 0
        for t in self.index_list:
            s += " Revolution %u: %.2fms\n" % (rev, t*1000/self.sample_freq)
            rev += 1
        return s[:-1]


    def summary_string(self):
        return ("Raw Flux (%u flux in %.2fms)"
                % (len(self.list), sum(self.list)*1000/self.sample_freq))


    def append(self, flux):
        # Scale the new flux if required, to match existing sample frequency.
        # This will result in floating-point flux values.
        if self.sample_freq == flux.sample_freq:
            f_list, i_list = flux.list, flux.index_list
        else:
            factor = self.sample_freq / flux.sample_freq
            f_list = [x*factor for x in flux.list]
            i_list = [x*factor for x in flux.index_list]
        # Any trailing flux is incorporated into the first revolution of
        # the appended flux.
        rev0 = i_list[0] + sum(self.list) - sum(self.index_list)
        self.index_list += [rev0] + i_list[1:]
        self.list += f_list


    def cue_at_index(self):

        if self.index_cued:
            return

        # Clip the initial partial revolution.
        to_index = self.index_list[0]
        for i in range(len(self.list)):
            to_index -= self.list[i]
            if to_index < 0:
                break
        if to_index < 0:
            self.list = [-to_index] + self.list[i+1:]
        else: # we ran out of flux
            self.list = []
        self.index_list = self.index_list[1:]
        self.index_cued = True


    def flux_for_writeout(self, cue_at_index=True):

        error.check(self.index_cued,
                    "Cannot write non-index-cued raw flux")
        error.check(self.splice == 0 or len(self.index_list) > 1,
                    "Cannot write single-revolution unaligned raw flux")
        splice_at_index = (self.splice == 0)

        # Copy the required amount of flux to a fresh list.
        flux_list = []
        to_index = self.index_list[0]
        remain = to_index + self.splice
        for f in self.list:
            if f > remain:
                break
            flux_list.append(f)
            remain -= f

        if not cue_at_index:
            # We will write more than one revolutionm and terminate the
            # second revolution at the splice. Extend the start of the write
            # with "safe" 4us sample values, in case the drive motor is a
            # little fast.
            if remain > 0:
                flux_list.append(remain)
            prepend = max(round(to_index/10 - self.splice), 0)
            if prepend != 0:
                four_us = max(self.sample_freq * 4e-6, 1)
                flux_list = [four_us]*round(prepend/four_us) + flux_list
            splice_at_index = False
        elif splice_at_index:
            # Extend with "safe" 4us sample values, to avoid unformatted area
            # at end of track if drive motor is a little slow.
            four_us = max(self.sample_freq * 4e-6, 1)
            if remain > four_us:
                flux_list.append(remain)
            for i in range(round(to_index/(10*four_us))):
                flux_list.append(four_us)
        elif remain > 0:
            # End the write exactly where specified.
            flux_list.append(remain)

        return WriteoutFlux(to_index, flux_list, self.sample_freq,
                            index_cued = cue_at_index,
                            terminate_at_index = splice_at_index)



    def flux(self):
        return self


    def scale(self, factor):
        """Scale up all flux and index timings by specified factor."""
        self.sample_freq /= factor


    @property
    def ticks_per_rev(self):
        """Mean time between index pulses, in sample ticks"""
        try:
            index_list = self.index_list
            if not self.index_cued:
                index_list = index_list[1:]
            ticks_per_rev = sum(index_list) / len(index_list)
        except:
            ticks_per_rev = self._ticks_per_rev
        return ticks_per_rev


    @property
    def time_per_rev(self):
        """Mean time between index pulses, in seconds (float)"""
        return self.ticks_per_rev / self.sample_freq


class WriteoutFlux(Flux):

    def __init__(self, ticks_to_index, flux_list, sample_freq,
                 index_cued, terminate_at_index):
        super().__init__([ticks_to_index], flux_list, sample_freq)
        self.index_cued = index_cued
        self.terminate_at_index = terminate_at_index


    def __str__(self):
        s = ("\nWriteoutFlux: %.2f MHz, %.2fms to index, %s\n"
             " Total: %u samples, %.2fms"
             % (self.sample_freq*1e-6,
                self.index_list[0]*1000/self.sample_freq,
                ("Write all", "Terminate at index")[self.terminate_at_index],
                len(self.list), sum(self.list)*1000/self.sample_freq))
        return s


    def summary_string(self):
        s = ("Flux: %.1fms period, %.1f ms total, %s"
             % (self.index_list[0]*1000/self.sample_freq,
                sum(self.list)*1000/self.sample_freq,
                ("Write all", "Terminate at index")[self.terminate_at_index]))
        return s


    def flux_for_writeout(self, cue_at_index=True):
        raise error.Fatal("WriteoutFlux: flux_for_writeout is unsupported")
 

    @property
    def ticks_per_rev(self):
        """Mean time between index pulses, in sample ticks"""
        return sum(self.index_list) / len(self.index_list)


# Local variables:
# python-indent: 4
# End:



# greaseweazle/image/d81.py
#
# Written & released by Keir Fraser <keir.xen@gmail.com>
#
# This is free and unencumbered software released into the public domain.
# See the file COPYING for more details, or visit <http://unlicense.org>.

from .img import IMG

class D81(IMG):
    default_format = 'commodore.1581'
    sides_swapped = True

# Local variables:
# python-indent: 4
# End:



# greaseweazle/codec/amiga/amigados.py
#
# Written & released by Keir Fraser <keir.xen@gmail.com>
#
# This is free and unencumbered software released into the public domain.
# See the file COPYING for more details, or visit <http://unlicense.org>.

import struct
import itertools as it
from bitarray import bitarray

from ..track import MasterTrack, RawTrack

default_revs = 1.1

sync_bytes = b'\x44\x89\x44\x89'
sync = bitarray(endian='big')
sync.frombytes(sync_bytes)

bad_sector = b'-=[BAD SECTOR]=-' * 32

class AmigaDOS:

    time_per_rev = 0.2

    def __init__(self, cyl, head):
        self.tracknr = cyl*2 + head
        self.sector = [None] * self.nsec
        self.map = [None] * self.nsec

    def summary_string(self):
        nsec, nbad = self.nsec, self.nr_missing()
        s = "AmigaDOS (%d/%d sectors)" % (nsec - nbad, nsec)
        return s

    # private
    def exists(self, sec_id, togo):
        return ((self.sector[sec_id] is not None)
                or (self.map[self.nsec-togo] is not None))

    # private
    def add(self, sec_id, togo, label, data):
        assert not self.exists(sec_id, togo)
        self.sector[sec_id] = label, data
        self.map[self.nsec-togo] = sec_id

    def has_sec(self, sec_id):
        return self.sector[sec_id] is not None

    def nr_missing(self):
        return len([sec for sec in self.sector if sec is None])

    def get_adf_track(self):
        tdat = bytearray()
        for sec in self.sector:
            tdat += sec[1] if sec is not None else bad_sector
        return tdat

    def set_adf_track(self, tdat):
        totsize = self.nsec * 512
        if len(tdat) < totsize:
            tdat += bytes(totsize - len(tdat))
        self.map = list(range(self.nsec))
        for sec in self.map:
            self.sector[sec] = bytes(16), tdat[sec*512:(sec+1)*512]
        return totsize

    def flux(self, *args, **kwargs):
        return self.raw_track().flux(*args, **kwargs)


    def decode_raw(self, track):
        raw = RawTrack(time_per_rev = self.time_per_rev,
                       clock = self.clock, data = track)
        bits, _ = raw.get_all_data()

        for offs in bits.itersearch(sync):

            if self.nr_missing() == 0:
                break

            sec = bits[offs:offs+544*16].tobytes()
            if len(sec) != 1088:
                continue

            header = decode(sec[4:12])
            format, track, sec_id, togo = tuple(header)
            if format != 0xff or track != self.tracknr \
               or not(sec_id < self.nsec and 0 < togo <= self.nsec) \
               or self.exists(sec_id, togo):
                continue

            label = decode(sec[12:44])
            hsum, = struct.unpack('>I', decode(sec[44:52]))
            if hsum != checksum(header + label):
                continue

            dsum, = struct.unpack('>I', decode(sec[52:60]))
            data = decode(sec[60:1084])
            gap = decode(sec[1084:1088])
            if dsum != checksum(data):
                continue;

            self.add(sec_id, togo, label, data)


    def raw_track(self):

        # List of sector IDs missing from the sector map:
        missing = iter([x for x in range(self.nsec) if not x in self.map])
        # Sector map with the missing entries filled in:
        full_map = [next(missing) if x is None else x for x in self.map]

        # Post-index track gap.
        t = encode(bytes(128 * (self.nsec//11)))

        for nr, sec_id in zip(range(self.nsec), full_map):
            sector = self.sector[sec_id]
            label, data = (bytes(16), bad_sector) if sector is None else sector
            header = bytes([0xff, self.tracknr, sec_id, self.nsec-nr])
            t += sync_bytes
            t += encode(header)
            t += encode(label)
            t += encode(struct.pack('>I', checksum(header + label)))
            t += encode(struct.pack('>I', checksum(data)))
            t += encode(data)
            t += encode(bytes(2))

        # Add the pre-index gap.
        tlen = (int((self.time_per_rev / self.clock)) + 31) & ~31
        t += bytes(tlen//8-len(t))

        track = MasterTrack(
            bits = mfm_encode(t),
            time_per_rev = 0.2)
        track.verify = self
        track.verify_revs = default_revs
        return track


    def verify_track(self, flux):
        cyl = self.tracknr // 2
        head = self.tracknr & 1
        readback_track = self.decode_track(cyl, head, flux)
        return (readback_track.nr_missing() == 0
                and self.sector == readback_track.sector)

    @classmethod
    def decode_track(cls, cyl, head, track):
        ados = cls(cyl, head)
        ados.decode_raw(track)
        return ados


class AmigaDOS_DD(AmigaDOS):
    nsec = 11
    clock = 14/7093790

class AmigaDOS_HD(AmigaDOS):
    nsec = 22
    clock = AmigaDOS_DD.clock / 2


def mfm_encode(dat):
    y = 0
    out = bytearray()
    for x in dat:
        y = (y<<8) | x
        if (x & 0xaa) == 0:
            y |= ~((y>>1)|(y<<1)) & 0xaaaa
        y &= 255
        out.append(y)
    return bytes(out)
    

def encode(dat):
    return bytes(it.chain(map(lambda x: (x >> 1) & 0x55, dat),
                          map(lambda x: x & 0x55, dat)))


def decode(dat):
    length = len(dat)//2
    return bytes(map(lambda x, y: (x << 1 & 0xaa) | (y & 0x55),
                     it.islice(dat, 0, length),
                     it.islice(dat, length, None)))


def checksum(dat):
    csum = 0
    for i in range(0, len(dat), 4):
        csum ^= struct.unpack('>I', dat[i:i+4])[0]
    return (csum ^ (csum>>1)) & 0x55555555


# Local variables:
# python-indent: 4
# End:



# greaseweazle/tools/util.py
#
# Greaseweazle control script: Utility functions.
#
# Written & released by Keir Fraser <keir.xen@gmail.com>
#
# This is free and unencumbered software released into the public domain.
# See the file COPYING for more details, or visit <http://unlicense.org>.

import argparse, os, sys, struct, time, re, platform
import importlib
from collections import OrderedDict

from .. import error


class CmdlineHelpFormatter(argparse.ArgumentDefaultsHelpFormatter,
                           argparse.RawDescriptionHelpFormatter):
    def _get_help_string(self, action):
        help = action.help
        if '%no_default' in help:
            return help.replace('%no_default', '')
        if ('%(default)' in help
            or action.default is None
            or action.default is False
            or action.default is argparse.SUPPRESS):
            return help
        return help + ' (default: %(default)s)'


def range_str(l):
    if len(l) == 0:
        return '<none>'
    p, str = None, ''
    for i in l:
        if p is not None and i == p[1]+1:
            p = p[0], i
            continue
        if p is not None:
            str += ('%d,' % p[0]) if p[0] == p[1] else ('%d-%d,' % p)
        p = (i,i)
    if p is not None:
        str += ('%d' % p[0]) if p[0] == p[1] else ('%d-%d' % p)
    return str

class TrackSet:

    class TrackIter:
        """Iterate over a TrackSet in physical <cyl,head> order."""
        def __init__(self, ts):
            l = []
            for c in ts.cyls:
                for h in ts.heads:
                    pc = c//-ts.step if ts.step < 0 else c*ts.step
                    pc += ts.h_off[h]
                    ph = 1-h if ts.hswap else h
                    l.append((pc, ph, c, h))
            l.sort()
            self.l = iter(l)
        def __next__(self):
            (self.physical_cyl, self.physical_head,
             self.cyl, self.head) = next(self.l)
            return self
    
    def __init__(self, trackspec):
        self.cyls = list()
        self.heads = list()
        self.h_off = [0]*2
        self.step = 1
        self.hswap = False
        self.trackspec = ''
        self.update_from_trackspec(trackspec)

    def update_from_trackspec(self, trackspec):
        """Update a TrackSet based on a trackspec."""
        self.trackspec += trackspec
        for x in trackspec.split(':'):
            if x == 'hswap':
                self.hswap = True
                continue
            k,v = x.split('=')
            if k == 'c':
                cyls = [False]*100
                for crange in v.split(','):
                    m = re.match('(\d\d?)(-(\d\d?)(/(\d))?)?$', crange)
                    if m is None: raise ValueError()
                    if m.group(3) is None:
                        s,e,step = int(m.group(1)), int(m.group(1)), 1
                    else:
                        s,e,step = int(m.group(1)), int(m.group(3)), 1
                        if m.group(5) is not None:
                            step = int(m.group(5))
                    for c in range(s, e+1, step):
                        cyls[c] = True
                self.cyls = []
                for c in range(len(cyls)):
                    if cyls[c]: self.cyls.append(c)
            elif k == 'h':
                heads = [False]*2
                for hrange in v.split(','):
                    m = re.match('([01])(-([01]))?$', hrange)
                    if m is None: raise ValueError()
                    if m.group(3) is None:
                        s,e = int(m.group(1)), int(m.group(1))
                    else:
                        s,e = int(m.group(1)), int(m.group(3))
                    for h in range(s, e+1):
                        heads[h] = True
                self.heads = []
                for h in range(len(heads)):
                    if heads[h]: self.heads.append(h)
            elif re.match('h[01].off$', k):
                h = int(re.match('h([01]).off$', k).group(1))
                m = re.match('([+-][\d])$', v)
                if m is None: raise ValueError()
                self.h_off[h] = int(m.group(1))
            elif k == 'step':
                m = re.match('1/(\d)$', v)
                self.step = -int(m.group(1)) if m is not None else int(v)
            else:
                print(k,v)
                raise ValueError()
        
    def __str__(self):
        s = 'c=%s' % range_str(self.cyls)
        s += ':h=%s' % range_str(self.heads)
        for i in range(len(self.h_off)):
            x = self.h_off[i]
            if x != 0:
                s += ':h%d.off=%s%d' % (i, '+' if x >= 0 else '', x)
        if self.step != 1:
            s += ':step=' + (('1/%d' % -self.step) if self.step < 0
                             else ('%d' % self.step))
        if self.hswap: s += ':hswap'
        return s

    def __iter__(self):
        return self.TrackIter(self)


image_types = OrderedDict(
    { '.adf': 'ADF',
      '.ads': ('ADS','acorn'),
      '.adm': ('ADM','acorn'),
      '.adl': ('ADL','acorn'),
      '.d81': 'D81',
      '.dsd': ('DSD','acorn'),
      '.dsk': 'EDSK',
      '.hfe': 'HFE',
      '.ima': 'IMG',
      '.img': 'IMG',
      '.ipf': 'IPF',
      '.raw': 'KryoFlux',
      '.sf7': 'SF7',
      '.scp': 'SCP',
      '.ssd': ('SSD','acorn'),
      '.st' : 'IMG' })

def get_image_class(name):
    if os.path.isdir(name):
        typespec = 'KryoFlux'
    else:
        _, ext = os.path.splitext(name)
        error.check(ext.lower() in image_types,
                    """\
%s: Unrecognised file suffix '%s'
Known suffixes: %s"""
                    % (name, ext, ', '.join(image_types)))
        typespec = image_types[ext.lower()]
    if isinstance(typespec, tuple):
        typename, classname = typespec
    else:
        typename, classname = typespec, typespec.lower()
    mod = importlib.import_module('fluxvis.greaseweazle.image.' + classname)
    return mod.__dict__[typename]


def with_drive_selected(fn, usb, args, *_args, **_kwargs):
    usb.set_bus_type(args.drive[0])
    try:
        usb.drive_select(args.drive[1])
        usb.drive_motor(args.drive[1], _kwargs.pop('motor', True))
        fn(usb, args, *_args, **_kwargs)
    except KeyboardInterrupt:
        print()
        usb.reset()
        raise
    finally:
        usb.drive_motor(args.drive[1], False)
        usb.drive_deselect()


def valid_ser_id(ser_id):
    return ser_id and ser_id.upper().startswith("GW")



# Local variables:
# python-indent: 4
# End:



# greaseweazle/image/kryoflux.py
#
# Written & released by Keir Fraser <keir.xen@gmail.com>
#
# This is free and unencumbered software released into the public domain.
# See the file COPYING for more details, or visit <http://unlicense.org>.

import struct, re, math, os
import itertools as it

from .. import error
from ..flux import Flux
from .image import Image

mck = 18432000 * 73 / 14 / 2
sck = mck / 2

class Op:
    Nop1  =  8
    Nop2  =  9
    Nop3  = 10
    Ovl16 = 11
    Flux3 = 12
    OOB   = 13

class OOB:
    StreamInfo =  1
    Index      =  2
    StreamEnd  =  3
    KFInfo     =  4
    EOF        = 13

class KryoFlux(Image):

    def __init__(self, name):
        if os.path.isdir(name):
            self.basename = os.path.join(name, '')
        else:
            m = re.search("(\d{2}.[01])?.raw$", name)
            self.basename = name[:m.start()]


    @classmethod
    def to_file(cls, name, fmt, noclobber):
        kf = cls(name)
        kf.noclobber = noclobber
        return kf

    @classmethod
    def from_file(cls, name):
        return cls(name)


    def get_track(self, cyl, side):

        name = self.basename + '%02d.%d.raw' % (cyl, side)
        try:
            with open(name, 'rb') as f:
                dat = f.read()
        except FileNotFoundError:
            return None

        # Parse the index-pulse stream positions.
        index = []
        idx = 0
        while idx < len(dat):
            op = dat[idx]
            if op == Op.OOB:
                oob_op, oob_sz = struct.unpack('<BH', dat[idx+1:idx+4])
                idx += 4
                if oob_op == OOB.Index:
                    pos, = struct.unpack('<I', dat[idx:idx+4])
                    index.append(pos)
                elif oob_op == OOB.EOF:
                    break
                idx += oob_sz
            elif op == Op.Nop3 or op == Op.Flux3:
                idx += 3
            elif op <= 7 or op == Op.Nop2:
                idx += 2
            else:
                idx += 1

        # Build the flux and index lists for the Flux object.
        flux, flux_list, index_list = [], [], []
        val, index_idx, stream_idx, idx = 0, 0, 0, 0
        while idx < len(dat):
            if index_idx < len(index) and stream_idx >= index[index_idx]:
                # We've passed an index marker.
                index_list.append(sum(flux))
                flux_list += flux
                flux = []
                index_idx += 1
            op = dat[idx]
            if op <= 7:
                # Flux2
                val += (op << 8) + dat[idx+1]
                flux.append(val)
                val = 0
                stream_idx += 2
                idx += 2
            elif op <= 10:
                # Nop1, Nop2, Nop3
                nr = op-7
                stream_idx += nr
                idx += nr
            elif op == Op.Ovl16:
                # Ovl16
                val += 0x10000
                stream_idx += 1
                idx += 1
            elif op == Op.Flux3:
                # Flux3
                val += (dat[idx+1] << 8) + dat[idx+2]
                flux.append(val)
                val = 0
                stream_idx += 3
                idx += 3
            elif op == Op.OOB:
                # OOB
                oob_op, oob_sz = struct.unpack('<BH', dat[idx+1:idx+4])
                idx += 4
                if oob_op == OOB.StreamInfo or oob_op == OOB.StreamEnd:
                    pos, = struct.unpack('<I', dat[idx:idx+4])
                    error.check(pos == stream_idx,
                                "Out-of-sync during KryoFlux stream read")
                elif oob_op == OOB.EOF:
                    break
                idx += oob_sz
            else:
                # Flux1
                val += op
                flux.append(val)
                val = 0
                stream_idx += 1
                idx += 1

        flux_list += flux

        # Crop partial first revolution.
        if len(index_list) > 1:
            short_index, index_list = index_list[0], index_list[1:]
            flux = 0
            for i in range(len(flux_list)):
                if flux >= short_index:
                    break
                flux += flux_list[i]
            flux_list = flux_list[i:]

        return Flux(index_list, flux_list, sck)


    def emit_track(self, cyl, side, track):
        """Converts @track into a KryoFlux stream file."""

        # Check if we should insert an OOB record for the next index mark.
        def check_index(prev_flux):
            nonlocal index_idx, dat
            if index_idx < len(index) and total >= index[index_idx]:
                dat += struct.pack('<2BH3I', Op.OOB, OOB.Index, 12,
                                   stream_idx,
                                   round(index[index_idx] - total + prev_flux),
                                   round(index[index_idx]/8))
                index_idx += 1

        # Emit a resampled flux value to the KryoFlux data stream.
        def emit(f):
            nonlocal stream_idx, dat, total
            while f >= 0x10000:
                stream_idx += 1
                dat.append(Op.Ovl16)
                f -= 0x10000
                total += 0x10000
                check_index(0x10000)
            if f >= 0x800:
                stream_idx += 3
                dat += struct.pack('>BH', Op.Flux3, f)
            elif Op.OOB < f < 0x100:
                stream_idx += 1
                dat.append(f)
            else:
                stream_idx += 2
                dat += struct.pack('>H', f)
            total += f
            check_index(f)

        flux = track.flux()

        # HxC crashes or fails to load non-index-cued stream files.
        # So let's give it what it wants.
        flux.cue_at_index()

        factor = sck / flux.sample_freq
        dat = bytearray()

        # Start the data stream with a dummy index if our Flux is index cued.
        if flux.index_cued:
            dat += struct.pack('<2BH3I', Op.OOB, OOB.Index, 12, 0, 0, 0)

        # Prefix-sum list of resampled index timings.
        index = list(it.accumulate(map(lambda x: x*factor, flux.index_list)))
        index_idx = 0
        
        stream_idx, total, rem = 0, 0, 0.0
        for x in flux.list:
            y = x * factor + rem
            f = round(y)
            rem = y - f
            emit(f)

        # We may not have enough flux to get to the final index value.
        # Generate a dummy flux just enough to get us there.
        if index_idx < len(index):
            emit(math.ceil(index[index_idx] - total) + 1)
        # A dummy cell so that we definitely have *something* after the
        # final OOB.Index, so that all parsers should register the Index.
        emit(round(sck*12e-6)) # 12us

        # Emit StreamEnd and EOF blocks to terminate the stream.
        dat += struct.pack('<2BH2I', Op.OOB, OOB.StreamEnd, 8, stream_idx, 0)
        dat += struct.pack('<2BH', Op.OOB, OOB.EOF, 0x0d0d)

        name = self.basename + '%02d.%d.raw' % (cyl, side)
        with open(name, ('wb','xb')[self.noclobber]) as f:
                f.write(dat)


    def __enter__(self):
        return self
    def __exit__(self, type, value, tb):
        pass


# Local variables:
# python-indent: 4
# End:



# SPDX-FileCopyrightText: 2022 Jeff Epler for Adafruit Industries
#
# SPDX-License-Identifier: MIT

# pylint: disable=consider-using-f-string

"""Compute an inverse polar transform

This can warp an image from linear to polar.

It is not quite the inverse of the skimage polar mapping, but it performs the
task needed for fluxvis
"""

import numpy as np

from skimage import data
from skimage.transform import warp


def _inverse_polar_mapping(
    output_coords, k_angle, k_radius, center
):  #  pylint: disable=invalid-name
    """Compute an inverse polar mapping"""
    yy = output_coords[:, 0] - center[0]
    xx = output_coords[:, 1] - center[1]
    cc = (np.pi + np.arctan2(-yy, xx)) * k_angle
    rr = np.hypot(yy, xx) * k_radius
    return np.column_stack((cc, rr))


def warp_inverse_polar(
    image, center=None, *, radius=None, output_shape=None, multichannel=False, **kwargs
):
    """Warp an image from linear to polar"""
    if image.ndim != 2 and not multichannel:
        raise ValueError(
            "Input array must be 2 dimensions "
            "when `multichannel=False`,"
            " got {}".format(image.ndim)
        )

    if image.ndim != 3 and multichannel:
        raise ValueError(
            "Input array must be 3 dimensions "
            "when `multichannel=True`,"
            " got {}".format(image.ndim)
        )

    if output_shape is None:
        output_shape = np.array((960, 960))
    else:
        output_shape = np.array(output_shape[:2])

    input_shape = np.array(image.shape)[:2]
    if center is None:
        center = (output_shape[:2] / 2) - 0.5

    if radius is None:
        width, height = output_shape[:2] / 2
        radius = np.hypot(width, height) / np.sqrt(2)

    height = input_shape[0]
    width = input_shape[1]

    k_radius = height / radius
    k_angle = (width - 1) / (2 * np.pi)

    warp_args = {"k_angle": k_angle, "k_radius": k_radius, "center": center}
    warped = warp(
        image,
        _inverse_polar_mapping,
        map_args=warp_args,
        output_shape=output_shape,
        **kwargs
    )

    return warped


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    checkerboard = data.checkerboard()

    fig, axes = plt.subplots(1, 2, figsize=(8, 8))
    ax = axes.ravel()
    ax[0].set_title("Original")
    ax[0].imshow(checkerboard)
    ax[1].set_title("Original, Inverse-Polar-Transformed")
    ax[1].imshow(warp_inverse_polar(checkerboard))
    plt.show()
